//*****************************************************************************
//*
//*
//*     SocketUtility.cpp
//*
//*
//*****************************************************************************
//
//  This is a UDP/TCP/EVENT library with callback support for
//
//      LWIP
//      LINUX
//      WINDOWS
//		FREERTOS
//      RT-TARGET
//
//	Global Macros:
//
//		GLOBAL_CONFIG		Enable the configuration via GlobalConfig.h header file (1=on 0=off)
//		SU_USE_IPV6			Enable the IP-V6 support (0=disable 1=enable)
//		SU_HAS_SUBCALLS		Enable the callback support (0=off 1..3=levels)
//		NO_DYNAMIC_MEM		Use static memory for TCP handles (0=no 1=yes)
//		SOCKETS_MAX			Maximum count of static TCP handles (only if NO_DYNAMIC_MEM)
//		RECV_TIMEOUT		Default receive  timeout in seconds
//		SEND_TIMEOUT		Default transmit timeout in seconds
//		SAE_API				Is the API prefix for the functions
//		LWIP				Use the LWIP stack (only for RTTARGET,FREERTOS and WINDOWS)
//		FREERTOS			Operating system is FreeRTOS
//		RTTARGET			Operating system is RtTarget
//
//  Copyright  2018    Anton Zechner
//
//
#ifdef      GLOBAL_CONFIG
#define     SOCKETUTILITY_CPP
#include    <GlobalConfig.h>
#undef      SOCKETUTILITY_CPP
#endif

#include    <stdio.h>
#include    <stdlib.h>
#include    "SocketUtility.h"

#ifndef     NO_DYNAMIC_MEM
#define     NO_DYNAMIC_MEM      0
#endif

#ifndef     RECV_TIMEOUT
#define     RECV_TIMEOUT        0
#endif

#ifndef     SEND_TIMEOUT
#define     SEND_TIMEOUT        15
#endif

#ifndef 	SUB_BUFFER
#define 	SUB_BUFFER			1024
#endif

#ifndef     SAE_API
#define     SAE_API             extern
#endif

#if         SU_HAS_SUBCALLS>3
#error      wrong value vor SU_HAS_SUBCALLS
#endif

#ifdef      _MSC_VER
#pragma     warning( disable :4035 )
#pragma     warning( disable :4996 )
#endif

#define     SUAPI               extern "C"
#define     FLAG_BROKEN         0x00000001
#define     FLAG_ACCEPT         0x00000002
#define     FLAG_WRITE          0x00000004
#define     FLAG_DOWN           0x00000010
#define     LOOPBACK            (*(unsigned*)"\x7F\0\0\1")
#define     IPV4TOV6            "\0\0\0\0\0\0\0\0\0\0\xFF\xFF\0\0\0\0"
#define     IS_TAG_CLIENT(h)    (((char*)(h))==TAG_TCP)
#define     IS_TAG_SERVER(h)    (((char*)(h))>=TAG_SERVER)
#define     IS_TAG_MULTI(h)     (((char*)(h))> TAG_SERVER)
#define     IS_TAG_EVENT(h)     (((char*)(h))==TAG_EVENT)
#define     IS_TAG_UDP(h)       (((char*)(h))==TAG_UDP)
#define     IS_TAG_TCP(h)       (((char*)(h))>=TAG_TCP)
#define     TAG_SERVER          ((char*)0xFFFF)
#define     TAG_TCP             ((char*)0xFFFE)
#define     TAG_UDP             ((char*)0xFFFD)
#define     TAG_EVENT           ((char*)0xFFFC)


static char cHostName[64] = "";

#if         SU_USE_IPV6
static int  iIpV6Allowed  = SU_USE_IPV6;
static int  iDualStack    = SU_USE_IPV6;
#endif



//*****************************************************************************
//*
//*     WINDOWS
//*
//*****************************************************************************
#if defined(_WIN32) && !defined(RTTARGET) && !defined(LWIP)

    #pragma     comment(lib,"wsock32.lib")
    #pragma     comment(lib,"ws2_32.lib")
    #include    <winsock2.h>
    #include    <windows.h>
    #include    <malloc.h>

    #if defined(_M_X64) ||  defined(_M_AMD64)
    #pragma     warning (disable: 4244)
    #endif

    #if         defined(_DEBUG) && !defined(NO_DEBUG_NEW) && defined(_MSC_VER)
    extern "C"  void * _cdecl               _malloc_dbg(size_t iSize,int iType,const char *pFile,int iLine);
    extern "C"  void   _cdecl               _free_dbg  (void  *pData,int iType);
    #define     malloc(s)                   _malloc_dbg(s,1,__FILE__,__LINE__)
    #define     free(p)                     _free_dbg  (p,1)
    #endif

    #undef      FLAG_BROKEN
    #undef      FLAG_ACCEPT
    #undef      FLAG_WRITE
    #undef      FLAG_DOWN

    #define     FLAG_BROKEN                 FD_CLOSE
    #define     FLAG_ACCEPT                 FD_ACCEPT
    #define     FLAG_WRITE                  FD_WRITE
    #define     FLAG_DOWN                   0x80

    #define     SysSleep                    Sleep
    #define     SysTickCount                GetTickCount
    #define     LEN_TYPE                    int*
    #define     SysDW                       unsigned
    #define     SEND_FLAGS                  0
    #define     RECV_FLAGS                  0

    #define     WSAWaitForMultipleEvents(n,p,w,t,a) WaitForMultipleObjectsEx(n,p,w,t,a)
    #define     ioctlsocket(s,c,a)          WSAIoctl(s,c,a,sizeof(ULONG),a,sizeof(ULONG),&dwDummy,NULL,NULL)
    #define     WASResetEvent(e)            ResetEvent(e)
    #define     WASSetEvent(e)              SetEvent(e)

    #define     Semaphore                   LPCRITICAL_SECTION
    #define     SemaphoreExit(h)            {if(h){DeleteCriticalSection(h);free(h);}}
    #define     SemaphoreLock(h)            {if(h) EnterCriticalSection(h);}
    #define     SemaphoreUnlock(h)          {if(h) LeaveCriticalSection(h);}


    static inline LPCRITICAL_SECTION SemaphoreInit()
        {
        LPCRITICAL_SECTION  pLock;
        pLock = (LPCRITICAL_SECTION)malloc(sizeof(CRITICAL_SECTION));
        InitializeCriticalSection(pLock);
        return pLock;
        }

    static      Semaphore   hGlobalSem = 0;
    static      DWORD       dwDummy;

    typedef struct
        {
        HANDLE  hEvent;
        }SocketLocal;

    static inline int SocketAlloc(SocketLocal *pLocal)
        {
        pLocal->hEvent = WSACreateEvent();
        return 1;
        }

    static inline void SocketClose(SocketLocal *pLocal)
        {
        WSACloseEvent(pLocal->hEvent);
        }

    static inline void SocketSignal(HANDLE hEvent)
        {
        WSASetEvent(hEvent);
        }

    static inline void SocketReset(HANDLE hEvent)
        {
        WSAResetEvent(hEvent);
        }

    static int SocketWaitRd(int iTimeout,int iSocket,HANDLE hEvent,unsigned char volatile &ucConnFlags,unsigned uMin=1)
        {
        WSANETWORKEVENTS    sInfo;
        unsigned long       uByte;
        unsigned            uStop;
        int                 iRet;

        if(ucConnFlags&FLAG_BROKEN)
            {
            return 1;
            }

        if(!ioctlsocket(iSocket,FIONREAD,&uByte) && uByte>=uMin)
            {
            return 1;
            }

        if(iTimeout<=0)
            {
            return 0;
            }

        uStop = GetTickCount()+iTimeout;

        Retry:

           iRet = WSAWaitForMultipleEvents(1,&hEvent,FALSE,iTimeout,FALSE);
        if(iRet>WSA_WAIT_EVENT_0)return  0;
        if(iRet<WSA_WAIT_EVENT_0)return -1;

        if(!WSAEnumNetworkEvents(iSocket,NULL,&sInfo))
            {
               ucConnFlags |= (sInfo.lNetworkEvents&~FLAG_DOWN);
            if(ucConnFlags&FD_CLOSE)return 0;

            WSAResetEvent(hEvent);
            }
        else{
            return 0;
            }

        if(!(ucConnFlags&FD_READ) || ioctlsocket(iSocket,FIONREAD,&uByte) || uByte<uMin)
            {
               iTimeout = uStop-GetTickCount();
            if(iTimeout>0)goto Retry;

            return 0;
            }

        return 1;
        }

    static int SocketWaitWr(int iTimeout,int iSocket,HANDLE hEvent,unsigned char volatile &ucConnFlags)
        {
        WSANETWORKEVENTS    sInfo;
        unsigned            uStop;
        int                 iRet;

        if(ucConnFlags&FLAG_BROKEN)
            {
            return 1;
            }

        uStop = GetTickCount()+iTimeout;

        Retry:

           iRet = WSAWaitForMultipleEvents(1,&hEvent,FALSE,iTimeout,FALSE);
        if(iRet>WSA_WAIT_EVENT_0)return  0;
        if(iRet<WSA_WAIT_EVENT_0)return -1;

        if(!WSAEnumNetworkEvents(iSocket,NULL,&sInfo))
            {
               ucConnFlags |= (sInfo.lNetworkEvents&~FLAG_DOWN);
            if(ucConnFlags&FD_CLOSE)return 0;

            WSAResetEvent(hEvent);
            }
        else{
            return 0;
            }

        if(!(ucConnFlags&FD_WRITE))
            {
               iTimeout = uStop-GetTickCount();
            if(iTimeout>0)goto Retry;
            return 0;
            }

        return 1;
        }


    #if SU_USE_IPV6

    static int SocketWaitAc(int iTimeout,int iSocketA,HANDLE hEventA,unsigned char ucConnFlags,int iSocketB,HANDLE hEventB)
        {
        HANDLE              aEvents[2];
        WSANETWORKEVENTS    sInfo;
        int                 iRet;

        aEvents[0] = hEventA;
        aEvents[1] = hEventB;

           iRet = WSAWaitForMultipleEvents((hEventB)? 2:1,aEvents,FALSE,iTimeout,FALSE);
        if(iRet>WSA_WAIT_EVENT_0+1)return  0;
        if(iRet<WSA_WAIT_EVENT_0  )return -1;

        iRet -= WSA_WAIT_EVENT_0-1;

        if(!WSAEnumNetworkEvents((iRet==2)? iSocketB:iSocketA,NULL,&sInfo))
            {
            if(sInfo.lNetworkEvents&FD_CLOSE)
                {
                ucConnFlags |= FLAG_BROKEN;
                return 0;
                }

            if(sInfo.lNetworkEvents&FD_ACCEPT)
                {
                ucConnFlags |= FLAG_ACCEPT;
                }

            WSAResetEvent((iRet==2)? hEventB:hEventA);
            }

        return iRet-WSA_WAIT_EVENT_0+1;
        }

    #define     PARAM_DUAL(p)      ,(p)? p->iSocket:-1,(p)? p->hEvent:NULL

    #else

    #define     PARAM_DUAL(p)

    static int SocketWaitAc(int iTimeout,int iSocket,HANDLE hEvent,unsigned char &ucConnFlags)
        {
        WSANETWORKEVENTS    sInfo;
        int                 iRet;

        if(ucConnFlags&FLAG_BROKEN)
            {
            return 1;
            }

           iRet = WSAWaitForMultipleEvents(1,&hEvent,FALSE,iTimeout,FALSE);
        if(iRet>WSA_WAIT_EVENT_0)return  0;
        if(iRet<WSA_WAIT_EVENT_0)return -1;

        if(!WSAEnumNetworkEvents(iSocket,NULL,&sInfo))
            {
            if(sInfo.lNetworkEvents&FD_CLOSE)
                {
                ucConnFlags |= FLAG_BROKEN;
                return 0;
                }

            if(sInfo.lNetworkEvents&FD_ACCEPT)
                {
                ucConnFlags |= FLAG_ACCEPT;
                }

            WSAResetEvent(hEvent);
            }

        return 1;
        }

    #endif

    #define     SU_WINDOWS          1
    #define     SU_NO_SELECT_WAIT   1
    #define     PARAM_WAIT(p)       p->iSocket,p->hEvent,p->ucConnFlags
    #define     PARAM_SIG(p)        p->hEvent
    #define     LOCAL_SET(p,l)      p->hEvent=l.hEvent
    #define     LOCAL_GET(p,l)      l.hEvent=p->hEvent;p->hEvent=0


//*****************************************************************************
//*
//*     RTTARGET
//*
//*****************************************************************************
#elif defined(RTTARGET) && !defined(LWIP)

    #if         SU_USE_IPV6
    #error      "RTTARGET don't supports IPv6"
    #endif

    #define     DIAG_SECTION_KERNEL         DIAG_SECTION_DRIVER
    #define     AT_MOTHERBOARD              1
    #define     RTKERNEL                    1
    #define     RTKMSP3                     1
    #define     ONTMSP3                     1

    #include    <../Source/Rtip/RtIp.h>
    #include    <rttarget.h>
    #include    <rtipapi.h>
    #include    <socket.h>
    #include    <clock.h>

    #define     SysSleep(dwMillisec)        RTKDelay(CLKMilliSecsToTicks(dwMillisec))
    #define     SysTickCount()              CLKTicksToMilliSecs(RTKGetTime())
    #define     LEN_TYPE                    int*
    #define     SysDW                       unsigned
    #define     SEND_FLAGS                  0
    #define     RECV_FLAGS                  0

    #define     Semaphore                   RTKSemaphore
    #define     SemaphoreInit()             RTKOpenSemaphore(ST_COUNTING,1,0,0)
    #define     SemaphoreExit(h)            RTKDeleteSemaphore(&h)
    #define     SemaphoreLock(h)            RTKWait(h)
    #define     SemaphoreUnlock(h)          RTKSignal(h)

    static      Semaphore   hGlobalSem =0;

    typedef struct
        {
        int     iSocket;
        }SocketLocal;

    static inline int SocketAlloc(SocketLocal *pLocal)
        {
        int                     iSocket;
        static unsigned short   usTurn;
        struct sockaddr_in      sData;
        int                     iErr;
        int                     iVal;

           iSocket = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
        if(iSocket==INVALID_SOCKET)
            {
            return 0;
            }

        usTurn++;
        sData.sin_addr.s_un.S_addr  = 0xFE00007F|(usTurn<<8);
        sData.sin_family            = AF_INET;
        sData.sin_port              = 0xFEFD;

            iErr = bind(iSocket,(const struct sockaddr*)&sData,sizeof(sData));
        if(!iErr)
            {
            iVal                    = 1;
            sData.sin_port          = 0xFDFE;
            setsockopt(iSocket,SOL_SOCKET,SO_REUSEADDR,(const char*)&iVal,sizeof(iVal));
            iErr = connect(iSocket,(const struct sockaddr*)&sData,sizeof(sData));
            }

        if(iErr)
            {
            closesocket(iSocket);
            return 0;
            }

        pLocal->iSocket = iSocket;

        return 1;
        }

    static inline void SocketClose(SocketLocal *pLocal)
        {
        closesocket(pLocal->iSocket);
        }

    static inline void SocketSignal(int iSocket)
        {
        PANYPORT   pPort;

            pPort = api_sock_to_port(iSocket);
        if(!pPort)return;

        pPort->ctrl.rtip_errno = 1;
        os_set_read_signal(pPort,0);
        }

    static inline void SocketReset(int iSocket)
        {
        PANYPORT   pPort;

            pPort = api_sock_to_port(iSocket);
        if(!pPort)return;

        pPort->ctrl.rtip_errno = 0;
        os_set_read_signal(pPort,0);
        }

    #define     PARAM_DUAL(p) ,(p)? p->iSocket:-1
    #define     PARAM_WAIT(p)   p->iSocket
    #define     PARAM_SIG(p)    p->iSocket
    #define     LOCAL_SET(p,l)  p->iSocket=l.iSocket
    #define     LOCAL_GET(p,l)  l.iSocket=p->iSocket


//*****************************************************************************
//*
//*     LWIP mit RTTARGET
//*
//*****************************************************************************
#elif defined(RTTARGET) && defined(LWIP)

    #if         SU_USE_IPV6
    #error      "LWIP don't supports IPv6"
    #endif

    #ifndef     INVALID_SOCKET
    #define     INVALID_SOCKET  (~0)
    #endif

    #ifndef     SOCKET_ERROR
    #define     SOCKET_ERROR    (-1)
    #endif

    #define     LWIP_PROVIDE_NO_ERRNO


    #include    <rttarget.h>
    #include    <rtipapi.h>
    #include    <string.h>
    #include    <clock.h>

    #include    "../src/include/lwip/sockets.h"
    #include    "../src/include/lwip/netif.h"
    #include    "../src/include/lwip/arch.h"
    #include    "../src/include/lwip/tcp.h"

    #define     SysSleep(dwMillisec)        RTKDelay(CLKMilliSecsToTicks(dwMillisec))
    #define     SysTickCount()              CLKTicksToMilliSecs(RTKGetTime())
    #define     LEN_TYPE                    unsigned*
    #define     SysDW                       unsigned
    #define     SOCKET                      int
    #define     SEND_FLAGS                  0
    #define     RECV_FLAGS                  0
    #define     gethostbyname(a)            0

    #define     Semaphore                   RTKSemaphore
    #define     SemaphoreInit()             RTKOpenSemaphore(ST_COUNTING,1,0,0)
    #define     SemaphoreExit(h)            RTKDeleteSemaphore(&h)
    #define     SemaphoreLock(h)            RTKWait(h)
    #define     SemaphoreUnlock(h)          RTKSignal(h)

    static      Semaphore                   hGlobalSem =0;
    static      Semaphore                   hInfoSem   =0;

    struct hostent
        {
        char   *h_name;                     // official name of host
        char  **h_aliases;                  // alias list
        short   h_addrtype;                 // host address type
        short   h_length;                   // length of address
        char  **h_addr_list;                // list of addresses
        };

    typedef struct addrinfo_ex
        {
        int                 ai_flags;       // AI_PASSIVE, AI_CANONNAME, AI_NUMERICHOST
        int                 ai_family;      // PF_xxx
        int                 ai_socktype;    // SOCK_xxx
        int                 ai_protocol;    // 0 or IPPROTO_xxx for IPv4 and IPv6
        size_t              ai_addrlen;     // Length of ai_addr
        char               *ai_canonname;   // Canonical name for nodename
        struct sockaddr    *ai_addr;        // Binary address
        struct addrinfo_ex *ai_next;        // Next structure in linked list
        struct sockaddr     ai_adr;         // Binary address
        }addrinfo_ex_lwip;

    static addrinfo_ex_lwip aNetInfo[10];

    typedef struct
        {
        int     iSocket;
        }SocketLocal;

    static inline int SocketAlloc(SocketLocal *pLocal)
        {
        pLocal->iSocket = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);;
        return (pLocal->iSocket==INVALID_SOCKET)? 0:1;
        }

    static inline void SocketClose(SocketLocal *pLocal)
        {
        closesocket(pLocal->iSocket);
        }

    static inline void SocketSignal(int iSocket)
        {
        lwip_event_set(iSocket,1);
        }

    static inline void SocketReset(int iSocket)
        {
        lwip_event_set(iSocket,0);
        }

    #define     PARAM_DUAL(p) ,(p)? p->iSocket:-1
    #define     PARAM_WAIT(p)   p->iSocket
    #define     PARAM_SIG(p)    p->iSocket
    #define     LOCAL_SET(p,l)  p->iSocket=l.iSocket
    #define     LOCAL_GET(p,l)  l.iSocket=p->iSocket


//*****************************************************************************
//*
//*     LWIP mit FREERTOS
//*
//*****************************************************************************
#elif defined(FREERTOS) && defined(LWIP)

    #if         SU_USE_IPV6
    #error      "LWIP don't supports IPv6"
    #endif

    #ifndef     INVALID_SOCKET
    #define     INVALID_SOCKET  (~0)
    #endif

    #ifndef     SOCKET_ERROR
    #define     SOCKET_ERROR    (-1)
    #endif

    #define     LWIP_PROVIDE_NO_ERRNO

    #include    "FreeRTOS.h"
    #include    "iodefine.h"
    #include    "devdrv_riic.h"
    #include    "semphr.h"
    #include    "intc.h"

    #include    "../src/include/lwip/sockets.h"
    #include    "../src/include/lwip/netif.h"
    #include    "../src/include/lwip/arch.h"
    #include    "../src/include/lwip/tcp.h"

    #if         configTICK_RATE_HZ==1000
    #define     SYS_MS_TO_TICK(m)           (m)
    #define     SysTickCount()              xTaskGetTickCount()
    #else
    #define     SYS_MS_TO_TICK(m)           (((m)*configTICK_RATE_HZ+999)/1000)

    static      unsigned SysTickCount()
                {
                static unsigned uHigh[8]={0,0,0,0,0,0,0,0};
                static unsigned uLast=0;
                uint64          uTime;
                unsigned        uTick;
                unsigned        uNext;
                unsigned        uUpper;

                uTick   = xTaskGetTickCount();
                uUpper  = uTick>>29;

                if(uUpper!=uLast)
                    {
                    uLast  =  uUpper;
                    uNext  = (uUpper-4)&7;

                    if(uUpper>=4)
                        {
                        uHigh[uNext] = uHigh[uUpper]+1;
                        }
                    else{
                        uHigh[uNext] = uHigh[uUpper];
                        }
                    }

                uTime       = uHigh[uUpper];
                uTime     <<= 32;
                uTime      |= uTick;
                uTime      *= (0x10000*1000)/SYS_TICK_RATE_HZ;
                uTime      /=  0x10000;

                return (unsigned)uTime;
                }
    #endif

    #define     LEN_TYPE                    unsigned*
    #define     SysDW                       unsigned
    #define     SOCKET                      int
    #define     SEND_FLAGS                  0
    #define     RECV_FLAGS                  0
    #define     gethostbyname(a)            0
    #define     SysSleep(t)                 vTaskDelay(SYS_MS_TO_TICK(t))

    #define     Semaphore                   SemaphoreHandle_t
    #define     SemaphoreInit()             xSemaphoreCreateCounting(0x70000000,1)
    #define     SemaphoreExit(h)            vSemaphoreDelete(h)
    #define     SemaphoreLock(h)            xSemaphoreTake(h,1);
    #define     SemaphoreUnlock(h)          xSemaphoreGive(h)

    static      Semaphore                   hGlobalSem =0;
    static      Semaphore                   hInfoSem   =0;

    struct hostent
        {
        char   *h_name;                     // official name of host
        char  **h_aliases;                  // alias list
        short   h_addrtype;                 // host address type
        short   h_length;                   // length of address
        char  **h_addr_list;                // list of addresses
        };

    typedef struct addrinfo_ex
        {
        int                 ai_flags;       // AI_PASSIVE, AI_CANONNAME, AI_NUMERICHOST
        int                 ai_family;      // PF_xxx
        int                 ai_socktype;    // SOCK_xxx
        int                 ai_protocol;    // 0 or IPPROTO_xxx for IPv4 and IPv6
        size_t              ai_addrlen;     // Length of ai_addr
        char               *ai_canonname;   // Canonical name for nodename
        struct sockaddr    *ai_addr;        // Binary address
        struct addrinfo_ex *ai_next;        // Next structure in linked list
        struct sockaddr     ai_adr;         // Binary address
        }addrinfo_ex_lwip;

    static addrinfo_ex_lwip aNetInfo[10];


    typedef struct
        {
        int     iSocket;
        }SocketLocal;

    static inline int SocketAlloc(SocketLocal *pLocal)
        {
        pLocal->iSocket = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);;
        return (pLocal->iSocket==INVALID_SOCKET)? 0:1;
        }

    static inline void SocketClose(SocketLocal *pLocal)
        {
        closesocket(pLocal->iSocket);
        }

    static inline void SocketSignal(int iSocket)
        {
        lwip_event_set(iSocket,1);
        }

    static inline void SocketReset(int iSocket)
        {
        lwip_event_set(iSocket,0);
        }

    #define     PARAM_DUAL(p) ,(p)? p->iSocket:-1
    #define     PARAM_WAIT(p)   p->iSocket
    #define     PARAM_SIG(p)    p->iSocket
    #define     LOCAL_SET(p,l)  p->iSocket=l.iSocket
    #define     LOCAL_GET(p,l)  l.iSocket=p->iSocket



//*****************************************************************************
//*
//*     LWIP mit WINDOWS
//*
//*****************************************************************************
#elif defined(_WIN32) && defined(LWIP)

    #include    <string.h>
    #include    "../src/include/lwip/sockets.h"
    #include    "../src/include/lwip/netif.h"
    #include    "../src/include/lwip/arch.h"
    #include    "../src/include/lwip/tcp.h"
    #include    <malloc.h>

    #if         SU_USE_IPV6
    #error      "LWIP don't supports IPv6"
    #endif

    #if         defined(_DEBUG) && !defined(NO_DEBUG_NEW) && defined(_MSC_VER)
    extern "C"  void * _cdecl               _malloc_dbg(size_t iSize,int iType,const char *pFile,int iLine);
    extern "C"  void   _cdecl               _free_dbg  (void  *pData,int iType);
    #define     malloc(s)                   _malloc_dbg(s,1,__FILE__,__LINE__)
    #define     free(p)                     _free_dbg  (p,1)
    #endif

    #ifndef     INVALID_SOCKET
    #define     INVALID_SOCKET  (~0)
    #endif

    #ifndef     SOCKET_ERROR
    #define     SOCKET_ERROR    (-1)
    #endif

    #define     LWIP_PROVIDE_NO_ERRNO

    #define     SysSleep                    Sleep
    #define     SysTickCount                GetTickCount
    #define     LEN_TYPE                    unsigned*
    #define     SOCKET                      int
    #define     WINAPI                      __stdcall
    #define     SysDW                       unsigned
    #define     DWORD                       unsigned
    #define     HANDLE                      void*
    #define     LPVOID                      void*
    #define     SEND_FLAGS                  0
    #define     RECV_FLAGS                  0
    #define     gethostbyname(a)            0

    typedef struct _RTL_CRITICAL_SECTION
        {
        void       *DebugInfo;
        long        LockCount;
        long        RecursionCount;
        void       *OwningThread;
        void       *LockSemaphore;
        unsigned    SpinCount;
        }CRITICAL_SECTION, *LPCRITICAL_SECTION;

    typedef     DWORD                                  (__stdcall  *PTHREAD_START_ROUTINE)(LPVOID);
    extern "C"  void            __declspec(dllimport)   __stdcall   EnterCriticalSection(LPCRITICAL_SECTION pCriticalSection);
    extern "C"  void            __declspec(dllimport)   __stdcall   LeaveCriticalSection(LPCRITICAL_SECTION pCriticalSection);
    extern "C"  void            __declspec(dllimport)   __stdcall   DeleteCriticalSection(LPCRITICAL_SECTION pCriticalSection);
    extern "C"  void            __declspec(dllimport)   __stdcall   InitializeCriticalSection(LPCRITICAL_SECTION pCriticalSection);
    extern "C"  void            __declspec(dllimport) * __stdcall   CreateSemaphoreA(void*,long,long,const char*);
    extern "C"  unsigned long   __declspec(dllimport)   __stdcall   WaitForSingleObject(void *,unsigned long);
    extern "C"  unsigned long   __declspec(dllimport)   __stdcall   GetTickCount(void);
    extern "C"  int             __declspec(dllimport)   __stdcall   ReleaseSemaphore(void *,long,long*);
    extern "C"  int             __declspec(dllimport)   __stdcall   CloseHandle(void*);
    extern "C"  void            __declspec(dllimport)   __stdcall   Sleep(unsigned long);

    #define     Semaphore                   LPCRITICAL_SECTION
    #define     SemaphoreExit(h)            {if(h){DeleteCriticalSection(h);free(h);}}
    #define     SemaphoreLock(h)            {if(h) EnterCriticalSection(h);}
    #define     SemaphoreUnlock(h)          {if(h) LeaveCriticalSection(h);}

    static inline LPCRITICAL_SECTION SemaphoreInit()
        {
        LPCRITICAL_SECTION  pLock;
        pLock = (LPCRITICAL_SECTION)malloc(sizeof(CRITICAL_SECTION));
        InitializeCriticalSection(pLock);
        return pLock;
        }

    static      Semaphore                   hGlobalSem =0;
    static      Semaphore                   hInfoSem   =0;

    struct hostent
        {
        char   *h_name;                     // official name of host
        char  **h_aliases;                  // alias list
        short   h_addrtype;                 // host address type
        short   h_length;                   // length of address
        char  **h_addr_list;                // list of addresses
        };

    typedef struct addrinfo_ex
        {
        int                 ai_flags;       // AI_PASSIVE, AI_CANONNAME, AI_NUMERICHOST
        int                 ai_family;      // PF_xxx
        int                 ai_socktype;    // SOCK_xxx
        int                 ai_protocol;    // 0 or IPPROTO_xxx for IPv4 and IPv6
        size_t              ai_addrlen;     // Length of ai_addr
        char               *ai_canonname;   // Canonical name for nodename
        struct sockaddr    *ai_addr;        // Binary address
        struct addrinfo_ex *ai_next;        // Next structure in linked list
        struct sockaddr     ai_adr;         // Binary address
        }addrinfo_ex_lwip;

    static addrinfo_ex_lwip aNetInfo[10];


    typedef struct
        {
        int     iSocket;
        }SocketLocal;

    static inline int SocketAlloc(SocketLocal *pLocal)
        {
        pLocal->iSocket = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);;
        return (pLocal->iSocket==INVALID_SOCKET)? 0:1;
        }

    static inline void SocketClose(SocketLocal *pLocal)
        {
        closesocket(pLocal->iSocket);
        }

    static inline void SocketSignal(int iSocket)
        {
        lwip_event_set(iSocket,1);
        }

    static inline void SocketReset(int iSocket)
        {
        lwip_event_set(iSocket,0);
        }

    #define     PARAM_DUAL(p) ,(p)? p->iSocket:-1
    #define     PARAM_WAIT(p)   p->iSocket
    #define     PARAM_SIG(p)    p->iSocket
    #define     LOCAL_SET(p,l)  p->iSocket=l.iSocket
    #define     LOCAL_GET(p,l)  l.iSocket=p->iSocket



//*****************************************************************************
//*
//*     LINUX
//*
//*****************************************************************************
#elif defined(linux) || defined(LINUX)

    #ifndef     _GNU_SOURCE
    #define     _GNU_SOURCE
    #endif

    #include    <unistd.h>
    #include    <sys/stat.h>
    #include    <sys/time.h>
    #include    <sys/times.h>
    #include    <sys/types.h>
    #include    <sys/ioctl.h>
    #include    <sys/socket.h>
    #include    <linux/if_arp.h>
    #include    <netinet/tcp.h>
    #include    <netinet/in.h>
    #include    <arpa/inet.h>
    #include    <semaphore.h>
    #include    <pthread.h>
    #include    <ifaddrs.h>
    #include    <string.h>
    #include    <signal.h>
    #include    <malloc.h>
    #include    <fcntl.h>
    #include    <netdb.h>
    #include    <sched.h>
    #include    <time.h>


    #define     SOCKET  int

    #ifndef     LINUX
    #define     LINUX           (1)
    #endif

    #ifndef     SOCKET_ERROR
    #define     SOCKET_ERROR    (-1)
    #endif

    #ifndef     INVALID_SOCKET
    #define     INVALID_SOCKET  (~0)
    #endif

    #ifndef     TCP_NODELAY
    #define     TCP_NODELAY     (1)
    #endif

    #ifndef     TCP_QUICKACK
    #define     TCP_QUICKACK    (12)
    #endif

    #define     ioctlsocket                 ioctl
    #define     closesocket(Socket)         close(Socket)
    #define     LEN_TYPE                    socklen_t*
    #define     SysDW                       unsigned
    #define     SEND_FLAGS                  MSG_NOSIGNAL
    #define     RECV_FLAGS                  MSG_NOSIGNAL
    typedef     pthread_mutex_t            *Semaphore;


    extern int  sem_timedwait(sem_t*,const struct timespec*);

    static pthread_mutex_t *SemaphoreInit()
        {
        pthread_mutex_t *pSem;
        pSem = (pthread_mutex_t*)malloc(sizeof(pthread_mutex_t));
        pthread_mutex_init(pSem,0);
        return pSem;
        }

    static inline void SemaphoreExit(pthread_mutex_t *pSem)
        {
        if(!pSem)return;
        pthread_mutex_destroy(pSem);
        free(pSem);
        }

    static inline void SemaphoreLock(pthread_mutex_t *pSem)
        {
        pthread_mutex_lock(pSem);
        }

    static inline void SemaphoreUnlock(pthread_mutex_t *pSem)
        {
        pthread_mutex_unlock(pSem);
        }

    static void SysSleep(unsigned uMilliSec)
        {
        struct timeval      sTime;
        struct timespec     sSpec;
        pthread_mutex_t     sMutex;
        pthread_condattr_t  sAttr;
        pthread_cond_t      sCond;


        if(!clock_gettime(CLOCK_MONOTONIC,&sSpec))
            {
            pthread_condattr_init(&sAttr);
            pthread_condattr_setclock(&sAttr,CLOCK_MONOTONIC);
            pthread_cond_init(&sCond,&sAttr);

            sSpec.tv_nsec += (uMilliSec%1000)*1000000;
            sSpec.tv_sec  += (sSpec.tv_nsec )/1000000000;
            sSpec.tv_sec  += (uMilliSec/1000);
            sSpec.tv_nsec %= 1000000000;
            }
        else if(!clock_gettime(CLOCK_REALTIME,&sSpec))
            {
            pthread_condattr_init(&sAttr);
            pthread_condattr_setclock(&sAttr,CLOCK_REALTIME);
            pthread_cond_init(&sCond,&sAttr);

            sSpec.tv_nsec += (uMilliSec%1000)*1000000;
            sSpec.tv_sec  += (sSpec.tv_nsec )/1000000000;
            sSpec.tv_sec  += (uMilliSec/1000);
            sSpec.tv_nsec %= 1000000000;
            }
        else{
            gettimeofday(&sTime,0);

            sSpec.tv_nsec  =(sTime.tv_usec + (uMilliSec % 1000) * 1000) * 1000;
            sSpec.tv_sec   = sTime.tv_sec  + (uMilliSec / 1000) + (sSpec.tv_nsec / 1000000000);
            sSpec.tv_nsec %= 1000000000;
            }

        pthread_mutex_init    (&sMutex,0);
        pthread_cond_init     (&sCond ,0);

        pthread_mutex_lock    (&sMutex);
        pthread_cond_timedwait(&sCond, &sMutex, &sSpec);
        pthread_mutex_unlock  (&sMutex);

        pthread_cond_destroy  (&sCond );
        pthread_mutex_destroy (&sMutex);
        }

    static Semaphore        hGlobalSem = 0;

    static inline unsigned SysTickCount()
        {
        struct timespec     sTime;

        if(clock_gettime(CLOCK_MONOTONIC,&sTime))
            {
            clock_gettime(CLOCK_REALTIME,&sTime);
            }

        return sTime.tv_sec*1000+sTime.tv_nsec/1000000;
        }

    static int BindToInterface(int iSock,unsigned aAddr)
        {
        struct ifaddrs     *pList;
        struct ifaddrs     *pItem;
        struct ifreq        sInter;
        int                 iFamily;
        int                 iVal;
        int                 iRet;


        if(getifaddrs(&pList)==-1)
            {
            return 0;
            }


        iRet = -1;

        for(pItem=pList;pItem;pItem=pItem->ifa_next)                // enum all ip addresses
            {
            if(!pItem->ifa_name   )continue;
            if(!pItem->ifa_addr   )continue;

               iFamily = pItem->ifa_addr->sa_family;
            if(iFamily!=AF_INET)continue;


            if(aAddr!=*(unsigned*)(pItem->ifa_addr->sa_data+2))     // same address
                {
                continue;
                }

            memset(&sInter,0,sizeof(sInter));
            strncpy(sInter.ifr_name,pItem->ifa_name,sizeof(sInter.ifr_name)-1);

            if(setsockopt(iSock,SOL_SOCKET,SO_BINDTODEVICE,(void *)&sInter,sizeof(sInter))==0)
                {
                iRet = 1;
                iVal = 1;

                setsockopt(iSock,SOL_SOCKET,SO_REUSEADDR,(const char*)&iVal,sizeof(iVal));

                break;
                }

            }

        freeifaddrs(pList);


        return iRet;
        }

    typedef struct
        {
        int     aPipe[2];
        }SocketLocal;

    static inline int SocketAlloc(SocketLocal *pLocal)
        {
        int     iFlags;
        if(pipe(pLocal->aPipe))return 0;

        #ifdef F_SETPIPE_SZ
        int     iSize=8;
        fcntl(pLocal->aPipe[1],F_SETPIPE_SZ,&iSize);
        #endif

        iFlags = fcntl(pLocal->aPipe[0],F_GETFL);
        fcntl(pLocal->aPipe[0],F_SETFL,iFlags|O_NONBLOCK);
        iFlags = fcntl(pLocal->aPipe[1],F_GETFL);
        fcntl(pLocal->aPipe[1],F_SETFL,iFlags|O_NONBLOCK);
        return 1;
        }

    static inline void SocketClose(SocketLocal *pLocal)
        {
        close(pLocal->aPipe[0]);
        close(pLocal->aPipe[1]);
        }

    static inline void SocketSignal(int iSocket,int iPipe)
        {
        int     iBytes;
        int     iErr;

           iErr = ioctl(iSocket,FIONREAD,&iBytes);
        if(iErr<0 || iBytes>0)return;

        write(iPipe,"1",1);
        }

    static inline void SocketReset(int iSocket,int iPipe)
        {
        char    aBuffer[256];
        int     iBytes;
        int     iNum;
        int     iErr;

           iErr = ioctl(iSocket,FIONREAD,&iBytes);
        if(iErr<0 || iBytes<=0)return;

        while(iBytes>0)
            {
            iNum =(iBytes>256)? 256:iBytes;
            read(iSocket,aBuffer,iNum);
            iBytes -= iNum;
            }
        }

    #define     PARAM_DUAL(p) ,(p)? p->iSocket:-1
    #define     PARAM_WAIT(p)   p->iSocket
    #define     PARAM_SIG(p)    p->iSocket,p->iPipe
    #define     LOCAL_SET(p,l)  p->iSocket=l.aPipe[0];p->iPipe=l.aPipe[1]
    #define     LOCAL_GET(p,l)  l.aPipe[0]=p->iSocket;l.aPipe[1]=p->iPipe


#else

#error  wrong operating system

#endif


//*****************************************************************************
//*
//*     Handel definitions
//*
//*****************************************************************************
// REM AZ gegen header tauschen
#ifdef LINUX

#define sockaddr_in6_       sockaddr_in6

#else

struct in_addr6_
    {
    unsigned char           s6_addr[16];                        // IPv6 address
    };

struct sockaddr_in6_
    {
    short                   sin6_family;                        // AF_INET6
    unsigned short          sin6_port;                          // Transport level port number
    unsigned long           sin6_flowinfo;                      // IPv6 flow information
    struct in_addr6_        sin6_addr;                          // IPv6 address
    unsigned long           sin6_scope_id;                      // Set of interfaces for a scope.
    };

#endif

typedef struct
    {
    IpAddr                  sIpAddr;
    SOCKET                  iSocket;
    unsigned short          usPort;
    unsigned char           ucSockFlags;                        // see at UDP_...
    unsigned char           ucConnFlags;
    void                   *hDummy;

    #ifdef SU_WINDOWS
    WSAEVENT                hEvent;
    #endif

    #ifdef RTTARGET
    RTKSemaphore            hSemaphor;
    #endif

    #if SU_USE_IPV6
    union
    {
    struct sockaddr_in      sAddr;
    struct sockaddr_in6_    sAddr6;
    };
    #else
    struct sockaddr_in      sAddr;
    #endif
    }UdpData;

typedef struct
    {
    IpAddr                  sIpAddr;
    SOCKET                  iSocket;
    unsigned short          usPort;
    unsigned char           ucSockFlags;                        // see at TCP_...
    unsigned char           ucConnFlags;
    void                   *hServer;

    #ifdef SU_WINDOWS
    WSAEVENT                hEvent;
    #endif

    #if SU_USE_IPV6
    union
    {
    struct sockaddr_in      sAddr;
    struct sockaddr_in6_    sAddr6;
    };
    #else
    struct sockaddr_in      sAddr;
    #endif

    #if SU_HAS_SUBCALLS > 0
    TcpSubCalls            *pSubCalls;
	int						iSubLevels;
    #endif
    }TcpData;

typedef struct
    {
    IpAddr                  sIpAddr;
    SOCKET                  iSocket;
    unsigned short          usPort;
    unsigned char           ucSignale;
    unsigned char           ucConnFlags;
    void                   *hDummy;

    #ifdef SU_WINDOWS
    WSAEVENT                hEvent;
    #endif

    #ifdef LINUX
    int                     iPipe;
    #endif
    }EvnData;

#if SU_HAS_SUBCALLS
static int TcpSubCallRecv(TcpSubCalls *pSubCalls,int iLevels,int iBytes=-1);
static int TcpSubCallSend(TcpSubCalls *pSubCalls,int iLevels,int iBytes=-1,const char *pData=0);
#endif

//*****************************************************************************
//*
//*     Several Macros
//*
//*****************************************************************************

#if SU_USE_IPV6
#define     MULTICAST_V6            "\xFF\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"
#define     SET_V4_ADDR(p,a)        { (p)->aAddr[0] =SU_IPV4_ID;  (p)->aAddr[1]=(a);}
#define     GET_V4_ADDR(p)          (((p)->aAddr[0]==SU_IPV4_ID)? (p)->aAddr[1]:0   )
#else
#define     SET_V4_ADDR(p,a)        { (p)->aAddr[0]=(a);}
#define     GET_V4_ADDR(p)          ( (p)->aAddr[0]     )
#endif

#ifndef     IPPROTO_IPV6
#define     IPPROTO_IPV6            41
#endif

#ifndef     IPV6_V6ONLY
#define     IPV6_V6ONLY             27
#endif

#ifndef     AI_PASSIVE
#define     AI_PASSIVE              1
#endif



//*****************************************************************************
//*
//*     Use dynamic or static memory
//*
//*****************************************************************************

#if NO_DYNAMIC_MEM

#ifndef     SOCKETS_MAX
#define     SOCKETS_MAX     (128+SU_HAS_SUBCALLS*32)
#endif

#define     HANDLE_MAX      ((sizeof(TcpData)>sizeof(UdpData))? sizeof(TcpData):sizeof(UdpData))
#define     HANDLE_SIZE     ((HANDLE_MAX>sizeof(TcpSubCalls)*(SU_HAS_SUBCALLS+1))? HANDLE_MAX:sizeof(TcpSubCalls)*(SU_HAS_SUBCALLS+1))



static  Semaphore       hHandleSemaphore=0;
static  unsigned        uReserveFlags[(SOCKETS_MAX+31)/32];
static  char            cHandleMemory[ SOCKETS_MAX*HANDLE_SIZE];
SAE_API void            SocketAllocError(void);

inline void *MallocHandle(int)
{
unsigned    uBits;
int         i,j;



    if(!hHandleSemaphore)hHandleSemaphore=SemaphoreInit();

    SemaphoreLock(hHandleSemaphore);

    for(i=0;i<(SOCKETS_MAX+31)/32;i++)                      // find an empty slot
        {
           uBits=uReserveFlags[i];
        if(uBits==0xFFFFFFFF)continue;

        for(j=0;j<32;j++,uBits>>=1)
            {
            if(!(uBits&1))break;
            }

        if(i*32+j>=SOCKETS_MAX)break;

        uReserveFlags[i]|=1<<j;

        SemaphoreUnlock(hHandleSemaphore);

        return cHandleMemory+(i*32+j)*HANDLE_SIZE;
        }


    SemaphoreUnlock(hHandleSemaphore);


    #ifndef     NO_ALLOC_ERROR
    SocketAllocError();
    #endif



return 0;
}

inline void FreeHandle(void *pMemory)
{
unsigned uPos;


    uPos  = (char*)pMemory-cHandleMemory;
    uPos /= HANDLE_SIZE;

    if(uPos>=SOCKETS_MAX)return;


    SemaphoreLock(hHandleSemaphore);
    uReserveFlags[uPos/32] &= ~(1<<(uPos&0x1F));
    SemaphoreUnlock(hHandleSemaphore);

}

#else

#include    <malloc.h>

    inline void *MallocHandle(int iSize)  { return malloc(iSize);}
    inline void  FreeHandle(void *pMemory){ free(pMemory);}

#endif




//*****************************************************************************
//*
//*     InitSocket
//*
//*****************************************************************************
static  int             iSocketInit=0;
#define InitSocket()    if(!iSocketInit)InitSocketProc();
static  TcpHandle       TcpHandleAlloc(SOCKET iSocket,void *pData,int iServer,int iFlags);


#ifdef LWIP

static void InitSocketProc()
    {
    struct netif   *pInterface;
    unsigned        uPos;

    iSocketInit = 1;

    if(!hGlobalSem)hGlobalSem  = SemaphoreInit();
    if(!hInfoSem  )hInfoSem    = SemaphoreInit();

    #if LWIP_NETIF_HOSTNAME

    for(pInterface=netif_list;pInterface;pInterface=pInterface->next)
        {
        if(! pInterface->hostname)continue;
        if(!*pInterface->hostname)continue;

        IpSetHostname(pInterface->hostname);
        break;
        }

    #endif

    aNetInfo[0].ai_flags            = 0x04;
    aNetInfo[0].ai_family           = 2;
    aNetInfo[0].ai_socktype         = 0;
    aNetInfo[0].ai_protocol         = 0;
    aNetInfo[0].ai_addrlen          = 16;
    aNetInfo[0].ai_canonname        = 0;
    aNetInfo[0].ai_addr             = &aNetInfo[0].ai_adr;
    aNetInfo[0].ai_next             = 0;
    aNetInfo[0].ai_adr.sa_len       = 16;
    aNetInfo[0].ai_adr.sa_family    = 2;

    aNetInfo[1].ai_flags            = 0x04;
    aNetInfo[1].ai_family           = 2;
    aNetInfo[1].ai_socktype         = 0;
    aNetInfo[1].ai_protocol         = 0;
    aNetInfo[1].ai_addrlen          = 16;
    aNetInfo[1].ai_canonname        = 0;
    aNetInfo[1].ai_addr             = &aNetInfo[1].ai_adr;
    aNetInfo[1].ai_next             = 0;
    aNetInfo[1].ai_adr.sa_len       = 16;
    aNetInfo[1].ai_adr.sa_family    = 2;
    aNetInfo[1].ai_adr.sa_data[2]   = 127;
    aNetInfo[1].ai_adr.sa_data[3]   = 0;
    aNetInfo[1].ai_adr.sa_data[4]   = 0;
    aNetInfo[1].ai_adr.sa_data[5]   = 1;

    uPos = 2;

    for(pInterface=netif_list;pInterface;pInterface=pInterface->next)
        {
        if(!pInterface->ip_addr.addr)continue;

        *(unsigned*)(aNetInfo[uPos].ai_adr.sa_data+2) = pInterface->ip_addr.addr;

        aNetInfo[uPos].ai_flags         = 0x04;
        aNetInfo[uPos].ai_family        = 2;
        aNetInfo[uPos].ai_socktype      = 0;
        aNetInfo[uPos].ai_protocol      = 0;
        aNetInfo[uPos].ai_addrlen       = 16;
        aNetInfo[uPos].ai_canonname     = 0;
        aNetInfo[uPos].ai_addr          = &aNetInfo[uPos].ai_adr;
        aNetInfo[uPos].ai_next          =  aNetInfo+uPos+1;
        aNetInfo[uPos].ai_adr.sa_len    = 16;
        aNetInfo[uPos].ai_adr.sa_family = 2;

           uPos++;
        if(uPos>=8)break;
        }

    aNetInfo[uPos-1].ai_next = 0;
    }


#elif defined(SU_WINDOWS)

#ifndef SU_NO_EXIT
static void _cdecl ExitSocketProc()
    {
    WSACleanup();
    }
#endif

static void InitSocketProc()
    {
    WORD    wVersion;
    WSADATA sWsaData;


    wVersion = 0x0202;
    if(WSAStartup(wVersion,&sWsaData))
        {
        wVersion = 0x0002;
        WSAStartup(wVersion,&sWsaData);
        }

    iSocketInit = 1;
    hGlobalSem  = SemaphoreInit();

    #ifndef SU_NO_EXIT
    atexit(ExitSocketProc);
    #endif

    #if SU_USE_IPV6

    int     iSocket;

       iSocket = socket(AF_INET6,SOCK_DGRAM,IPPROTO_UDP);
    if(iSocket==INVALID_SOCKET)
        {
        iIpV6Allowed = 0;
        }
    else{
        closesocket(iSocket);
        }

    #endif
    }

#else

static void InitSocketProc()
    {
    iSocketInit = 1;
    hGlobalSem  = SemaphoreInit();


    #if SU_USE_IPV6

    int     iSocket;

       iSocket = socket(AF_INET6,SOCK_DGRAM,IPPROTO_UDP);
    if(iSocket==INVALID_SOCKET)
        {
        iIpV6Allowed = 0;
        }
    else{
        closesocket(iSocket);
        }

    #endif
    }

#endif

//*****************************************************************************
//*
//*     SocketWaitRd
//*
//*****************************************************************************
//  Waits for an receive or error event on on a socket.
//  iTimeout    : timeout in milli seconds
//  iSocket     : is the socket number
//  Returns the count of signaled sockets or -1 if an error occurs.
#ifndef SU_NO_SELECT_WAIT
static int SocketWaitRd(int iTimeout,SOCKET iSocket)
{
struct timeval  sTimeVal;
fd_set          sFdRead;
fd_set          sFdErr;
int             iRet;


    sTimeVal.tv_sec  =  iTimeout/1000;
    sTimeVal.tv_usec = (iTimeout%1000)*1000;

    FD_ZERO(&sFdRead);
    FD_ZERO(&sFdErr );

    FD_SET(iSocket,&sFdRead);
    FD_SET(iSocket,&sFdErr );

    iRet = select(iSocket+1,&sFdRead,0,&sFdErr,&sTimeVal);


return iRet;
}
#endif

//*****************************************************************************
//*
//*     SocketWaitWr
//*
//*****************************************************************************
//  Waits for an write possible or error event a socket.
//  iTimeout    : timeout in milli seconds
//  iSocketA    : first socket number
//  iSocketB    : second socket number (-1 means only one socket)
//  Returns the count of signaled sockets or -1 if an error occurs.
#ifndef SU_NO_SELECT_WAIT
static int SocketWaitWr(int iTimeout,SOCKET iSocket)
{
struct timeval  sTimeVal;
fd_set          sFdWrite;
int             iRet;


    sTimeVal.tv_sec  =  iTimeout/1000;
    sTimeVal.tv_usec = (iTimeout%1000)*1000;

    FD_ZERO(&sFdWrite);
    FD_SET(iSocket,&sFdWrite);

    iRet = select(iSocket+1,0,&sFdWrite,0,&sTimeVal);


return iRet;
}
#endif

//*****************************************************************************
//*
//*     SocketWaitAc
//*
//*****************************************************************************
//  Waits for an acczept event or error event on one or two sockets.
//  iTimeout    : timeout in milli seconds
//  iSocketA    : first socket number
//  iSocketB    : second socket number (-1 means only one socket)
//  Returns the the number of the signalled socket
//                  -1 = an error occurs
//                   0 = nothing signaled
//                   1 = first socket signaled
//                   2 = second socket signaled
#ifndef     SU_NO_SELECT_WAIT
#if         SU_USE_IPV6
#define     PARAM_IPV6      ,SOCKET iSocketB
#else
#define     PARAM_IPV6
#undef      PARAM_DUAL
#define     PARAM_DUAL(p)
#endif

static int SocketWaitAc(int iTimeout,SOCKET iSocketA PARAM_IPV6)
{
struct timeval  sTimeVal;
fd_set          sFdRead;
int             iMax;
int             iRet;


    sTimeVal.tv_sec  =  iTimeout/1000;
    sTimeVal.tv_usec = (iTimeout%1000)*1000;

    FD_ZERO(&sFdRead);
    FD_SET(iSocketA,&sFdRead);

    iMax = iSocketA+1;


    #if SU_USE_IPV6

    if(iSocketB<0)                                      // singel socket
        {
        iRet = select(iMax,&sFdRead,0,0,&sTimeVal);
        if(iRet>1)iRet=1;
        }
    else{                                               // dual wait for accept
        FD_SET(iSocketB,&sFdRead);

        if(iSocketB>iSocketA)iMax=iSocketB+1;

        iRet = select(iMax,&sFdRead,0,0,&sTimeVal);

        if(iRet>0)
            {
            iRet = FD_ISSET(iSocketB,&sFdRead)? 2:1;
            }
        }

    #else

        {
        iRet = select(iMax,&sFdRead,0,0,&sTimeVal);
        if(iRet>1)iRet=1;
        }

    #endif


return iRet;
}
#endif

//*****************************************************************************
//*
//*     getaddrinfo_ex
//*
//*****************************************************************************
//  Dummy fr Win200 weil hier getaddrinfo in der ws2_32.dll nicht existiert.

#ifdef LWIP


    void freeaddrinfo_ex(struct addrinfo_ex *pAddrInfo)
    {
        if(!pAddrInfo)return;
        SemaphoreUnlock(hInfoSem);
    }

    static int getaddrinfo_ex(const char *pNodeName,const char *pServname,const struct addrinfo_ex *pHints,struct addrinfo_ex **pResult)
    {
    unsigned char  *pData;
    unsigned        uNum;
    unsigned        uVal;


    if(!pNodeName || !*pNodeName)
        {
        *pResult = 0;
        return -1;
        }

    SemaphoreLock(hInfoSem);

    if(*pNodeName<='9' && *pNodeName>='0')
        {
        pData = (unsigned char*)aNetInfo[0].ai_adr.sa_data+2;

        for(uNum=0;uNum<4;uNum++)                       // decode ip address
            {
               uVal  = atoi(pNodeName);
            if(uVal>255)break;

            pData[uNum] = uVal;

            while(*pNodeName<='9' && *pNodeName>='0')
                {
                pNodeName++;
                }

            if(uNum==3)
                {
                if(*pNodeName!='\0')break;
                }
            else{
                if(*pNodeName!='.' )break;
                }

            pNodeName++;
            }

        if(uNum!=4)
            {
            SemaphoreUnlock(hInfoSem);
            *pResult = 0;
            return -1;
            }

        *pResult = aNetInfo+0;

        return 0;
        }

    if(*pNodeName=='l' && !strcmp(pNodeName,"localhost"))
        {
        *pResult = aNetInfo+1;

        return 0;
        }

    if(!strcmp(pNodeName,cHostName))
        {
        *pResult = aNetInfo+2;
        return 0;
        }

    SemaphoreUnlock(hInfoSem);
    *pResult = 0;
    return -1;
    }

#elif defined(SU_WINDOWS)


    typedef int (WSAAPI *getaddrinfo_type )(const char *,const char *,const struct addrinfo_ex *,struct addrinfo_ex **);
    typedef void(WSAAPI *freeaddrinfo_type)(struct addrinfo_ex *);

    typedef struct addrinfo_ex
        {
        int                 ai_flags;                   // AI_PASSIVE, AI_CANONNAME, AI_NUMERICHOST
        int                 ai_family;                  // PF_xxx
        int                 ai_socktype;                // SOCK_xxx
        int                 ai_protocol;                // 0 or IPPROTO_xxx for IPv4 and IPv6
        size_t              ai_addrlen;                 // Length of ai_addr
        char               *ai_canonname;               // Canonical name for nodename
        struct sockaddr    *ai_addr;                    // Binary address
        struct addrinfo_ex *ai_next;                    // Next structure in linked list
        }addrinfo_ex_;



    static  HMODULE             hDll    = 0;
    static  int                 iInit   = 0;
    static  getaddrinfo_type    pProc   = 0;
    static  freeaddrinfo_type   pFree   = 0;

    void freeaddrinfo_ex(struct addrinfo_ex *pAddrInfo)
    {
        if(!pFree)return;
        pFree(pAddrInfo);
    }

    static int getaddrinfo_ex(const char *pNodeName,const char *pServname,const struct addrinfo_ex *pHints,struct addrinfo_ex **pResult)
    {


        if(!iInit)
            {
            InitSocket();

            iInit = 1;

            if(!hDll)
                {
                    hDll = LoadLibrary("ws2_32.dll");
                if(!hDll)return -1;
                }

            if(!pProc)pProc = (getaddrinfo_type )GetProcAddress(hDll,"getaddrinfo" );
            if(!pFree)pFree = (freeaddrinfo_type)GetProcAddress(hDll,"freeaddrinfo");
            }


        if(!pProc)return -1;


    return pProc(pNodeName,pServname,pHints,pResult);
    }


#elif defined(RTTARGET) && !defined(LWIP)


    typedef struct addrinfo_ex
        {
        int                 ai_flags;                       // AI_PASSIVE, AI_CANONNAME, AI_NUMERICHOST
        int                 ai_family;                      // PF_xxx
        int                 ai_socktype;                    // SOCK_xxx
        int                 ai_protocol;                    // 0 or IPPROTO_xxx for IPv4 and IPv6
        size_t              ai_addrlen;                     // Length of ai_addr
        char               *ai_canonname;                   // Canonical name for nodename
        struct sockaddr    *ai_addr;                        // Binary address
        struct addrinfo_ex *ai_next;                        // Next structure in linked list
        }addrinfo_ex_;

    #define     getaddrinfo_ex(a,b,c,d) -1
    #define     freeaddrinfo_ex(a)


#elif defined(LWIP)


    typedef struct addrinfo_ex
        {
        int                 ai_flags;                       // AI_PASSIVE, AI_CANONNAME, AI_NUMERICHOST
        int                 ai_family;                      // PF_xxx
        int                 ai_socktype;                    // SOCK_xxx
        int                 ai_protocol;                    // 0 or IPPROTO_xxx for IPv4 and IPv6
        size_t              ai_addrlen;                     // Length of ai_addr
        char               *ai_canonname;                   // Canonical name for nodename
        struct sockaddr    *ai_addr;                        // Binary address
        struct addrinfo_ex *ai_next;                        // Next structure in linked list
        }addrinfo_ex_;

    #define     getaddrinfo_ex(a,b,c,d) -1
    #define     freeaddrinfo_ex(a)


#else

    // LINUX
    #define     freeaddrinfo_ex freeaddrinfo
    #define     getaddrinfo_ex  getaddrinfo
    #define     addrinfo_ex     addrinfo

#endif

//*****************************************************************************
//*
//*     CreateServerSocket
//*
//*****************************************************************************
//  Creates a server socked with automatic IPV4/IPV6 any address conversion
//  pAddrPtr    : pointer to the address pointer. It the address pointer
//                is zero, a any address is used.
//  pTemp       : temorary address object.
//  iMode       : socket mode  (0=tcp 1=udp)
//  Returns the socket or INVALID_SOCKET, and fills the address if the
//  address pointer is zero.
static SOCKET CreateServerSocket(IpAddr **pAddrPtr,IpAddr *pTemp,int iMode)
{
IpAddr     *pAddr=*pAddrPtr;
SOCKET      iSock;



    #if SU_USE_IPV6

    int iValue;

    if(!iIpV6Allowed)                                   // if no IPv6 possible use IPv4 any address
        {
        if(!pAddr)                                      // if zero then use 0.0.0.0
            {
            SET_V4_ADDR(pTemp,0);
            *pAddrPtr = pTemp;
            return INVALID_SOCKET;
            }

        if(pAddr->aAddr[0]==0 && pAddr->aAddr[1]==0 && pAddr->aAddr[2]==0)
            {
            if(pAddr->aAddr[3]==0)                      // use "0.0.0.0" instead of "::"
                {
                SET_V4_ADDR(pTemp,0);
                *pAddrPtr = pTemp;
                }
            else if(pAddr->aAddr[3]==1)                 // use "127.0.0.1" instead of "::1"
                {
                SET_V4_ADDR(pTemp,LOOPBACK);
                *pAddrPtr = pTemp;
                }
            }

        return INVALID_SOCKET;
        }

    if(iMode)                                           // is it udp
        {
        if(!pAddr)                                      // if zero then use 0.0.0.0
            {
            SET_V4_ADDR(pTemp,0);
            *pAddrPtr = pTemp;
            return INVALID_SOCKET;
            }

        if(pAddr->aAddr[0]==SU_IPV4_ID)                 // only IPV4 on udp
            {
            return INVALID_SOCKET;
            }
        }

    iSock  = socket(AF_INET6,(iMode)? SOCK_DGRAM:SOCK_STREAM,(iMode)? IPPROTO_UDP:IPPROTO_TCP);
    iValue = 0;                                         // enable dual-stack on the IPv6 socket

    if(setsockopt(iSock,IPPROTO_IPV6,IPV6_V6ONLY,(char*)&iValue,sizeof(iValue))==SOCKET_ERROR)
        {
        iDualStack = 0;

        if(!pAddr)
            {
            SET_V4_ADDR(pTemp,0);
            *pAddrPtr = pTemp;
            closesocket(iSock);
            return INVALID_SOCKET;
            }

        if(pAddr->aAddr[0]==SU_IPV4_ID)                 // if IPv4 don't use the IPv6 socket
            {
            closesocket(iSock);
            return INVALID_SOCKET;
            }

        return iSock;
        }


    iDualStack = 1;

                                                        // is it a any address
    if(!pAddr || (pAddr->aAddr[0]==SU_IPV4_ID && GET_V4_ADDR(pAddr)==0))
        {
        memset(pTemp,0,sizeof(IpAddr));                 // use "::" instead of "0.0.0.0"
        *pAddrPtr = pTemp;
        }

    #else

    if(!pAddr)                                          // use any address "0.0.0.0"
        {
        SET_V4_ADDR(pTemp,0);
        *pAddrPtr = pTemp;
        }

    iSock = INVALID_SOCKET;

    #endif


return iSock;
}

//*****************************************************************************
//*
//*     CreateTcpMulti
//*
//*****************************************************************************
//  Adds a IPv4 any listen port to an IPv6 any port
//  hBase       : is the IPv6 base handle
//  uPort       : is the uesed port number
//  iFlags      : Several Flags for the connection:
//                  TCP_NOBLOCKING = sets the socket in the no blocking mode
#if SU_USE_IPV6
static void CreateTcpMulti(TcpHandle hBase,unsigned short uPort,int iFlags)
{
unsigned long           lNoneBlocking=1;
struct sockaddr_in      sAddr;
TcpData                *pData;
TcpHandle               hNext;
SOCKET                  iSock;
int                     iErr;



       iSock=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
    if(iSock==INVALID_SOCKET)return;


    memset(&sAddr,0,sizeof(sAddr));

    sAddr.sin_family      =  AF_INET;
    sAddr.sin_port        =  htons(uPort);

        iErr = bind(iSock,(const struct sockaddr*)&sAddr,sizeof(sAddr));
    if(!iErr)iErr = listen(iSock,8);

    if(iErr)
        {
        closesocket(iSock);
        return;
        }

    if(iFlags&TCP_NOBLOCKING)                           // set socket in none blocking mode
        {
        lNoneBlocking = 1;

        if(ioctlsocket(iSock,FIONBIO,&lNoneBlocking))
            {
            closesocket(iSock);
            return;
            }
        }



        hNext = TcpHandleAlloc(iSock,&sAddr,TCP_SERVER,iFlags);
    if(!hNext)
        {
        closesocket(iSock);
        return;
        }

    pData = (TcpData*)hBase;
    pData->hServer  = hNext;


}
#endif

//*****************************************************************************
//*
//*     IpSetBcast
//*
//*****************************************************************************
//  Creates an IP V4 broadcast adddres
//  pAddr       : Poiter to the adress memory
//  uIpV4       : Is the 4 byte IP V4 address
//  uBits       : Is the count of network bits (0...31, 24 means 255.255.255.0)
//  Returns 0 if an error occurs or 1 ok.
int IpSetBcast(IpAddr *pAddr,unsigned uIpV4,unsigned uBits)
{

    if(!pAddr)return 0;

    if(uBits>=32)
        {
        SetIpV4(pAddr,0);
        return 0;
        }


    uBits  = 0xFFFFFFFF>>uBits;
    uIpV4 |= htonl(uBits);

    SetIpV4(pAddr,uIpV4);


return 1;
}

//*****************************************************************************
//*
//*     IpAddressA
//*
//*****************************************************************************
//  Converts an ip-string in an ip-address
//  pAddr       : Poiter to the adress memory
//  pIp         : IP-String  i.e. "127.0.0.1"
//  pMask       : Subnetmask for broadcast addresses (NULL for normal adresses)
//  Returns 0 if an error occurs or 1 if it is an IPV4 address or 2 for IPV6.
SUAPI int IpAddressA(IpAddr *pAddr,const char *pIp,const char *pMask)
{
unsigned    uIpAddr;
unsigned    uIpMask;




    if(!pIp || !*pIp)
        {
        SET_V4_ADDR(pAddr,INADDR_NONE);
        return 0;
        }

    if(!IpHostnameA(pAddr,pIp))
        {
        return 0;
        }

    #if SU_USE_IPV6

    if(pMask && IsIpV6(pAddr))
        {
        memcpy(pAddr->aAddr,MULTICAST_V6,16);
        return 2;
        }

    #endif

    if(pMask)
        {
           uIpMask = inet_addr(pMask);
        if(uIpMask==INADDR_NONE)
            {
            SET_V4_ADDR(pAddr,INADDR_NONE);
            return 0;
            }

        uIpAddr = GET_V4_ADDR(pAddr)|~uIpMask;

        SET_V4_ADDR(pAddr,uIpAddr);
        }



return 1;
}

//*****************************************************************************
//*
//*     IpAddressExA
//*
//*****************************************************************************
//  Converts an ip-string in an ip-address with pot
//  pAddr       : Poiter to the adress memory
//  pIp         : IP-String  i.e. "127.0.0.1:1234"
//                                "[::1]:1234"
//                                "www.google.de:1234"
//  pLen        : Here the size of the string after the port will be stored
//  Returns -1 if an error occurs or the port number.
SUAPI int IpAddressExA(IpAddr *pAddr,const char *pIp,int *pLen)
{
char    cName[64];
char   *pStop;
int     iPort;
int     iPos;


    if(!pIp)
        {
        if(pLen)*pLen=0;
        return -1;
        }

    if(*pIp=='[')                                       // IPV6 Name
        {
        for(iPos=0;iPos<63;iPos++)
            {
            if(pIp[iPos]== 0 )break;
            cName[iPos]=(char)pIp[iPos];
            if(pIp[iPos]==']'){iPos++;break;}
            }
        }
    else{
        for(iPos=0;iPos<63;iPos++)
            {
            if(pIp[iPos]== 0  )break;
            if(pIp[iPos]==':' )break;
            if(pIp[iPos]=='@' )break;
            if(pIp[iPos]=='#' )break;
            if(pIp[iPos]=='<' )break;
            if(pIp[iPos]=='>' )break;
            if(pIp[iPos]==';' )break;
            if(pIp[iPos]==',' )break;
            if(pIp[iPos]=='\t')break;
            cName[iPos]=(char)pIp[iPos];
            }
        }

    if(pIp[iPos]==':')                                  // exists a port address
        {
        cName[iPos]=0;
        iPort = strtol(pIp+iPos+1,&pStop,10);
        iPos  = pStop-pIp;
        }
    else{
        iPort = 0;
        }

    if(pLen)*pLen=iPos;
    cName[iPos]=0;

    if(!IpAddressA(pAddr,cName,0))
        {
        return -1;
        }


return iPort;
}


//*****************************************************************************
//*
//*     IpAddressW
//*
//*****************************************************************************
//  Converts an ip-string in an ip-address
//  pAddr       : Poiter to the adress memory
//  pIp         : IP-String  i.e. "127.0.0.1"
//  pMask       : Subnetmask for broadcast addresses (NULL for normal adresses)
//  Returns 0 if an error occurs or 1 if it is an IPV4 address or 2 for IPV6.
#if defined(WIN32)
SUAPI int IpAddressW(IpAddr *pAddr,const wchar_t *pIp,const wchar_t *pMask)
{
char    cName[64];
char    cMask[64];
int     iPos;


    if(pIp)
        {
        for(iPos=0;iPos<63;iPos++)
            {
            if(pIp[iPos]==0)break;
            cName[iPos]=(char)pIp[iPos];
            }

        cName[iPos]=0;
        }

    if(pMask)
        {
        for(iPos=0;iPos<63;iPos++)
            {
            if(pMask[iPos]==0)break;
            cMask[iPos]=(char)pMask[iPos];
            }

        cMask[iPos]=0;
        }

return IpAddressA(pAddr,(pIp)? cName:0,(pMask)? cMask:0);
}
#endif

//*****************************************************************************
//*
//*     IpAddressExW
//*
//*****************************************************************************
//  Converts an ip-string in an ip-address with pot
//  pAddr       : Poiter to the adress memory
//  pIp         : IP-String  i.e. "127.0.0.1:1234"
//                                "[::1]:1234"
//                                "www.google.de:1234"
//  pLen        : Here the size of the string after the port will be stored
//  Returns -1 if an error occurs or the port number.
#if defined(WIN32)
SUAPI int IpAddressExW(IpAddr *pAddr,const wchar_t *pIp,int *pLen)
{
char    cName[96];
int     iPos;


    if(!pIp)
        {
        if(pLen)*pLen=0;
        return -1;
        }

    for(iPos=0;iPos<95;iPos++)
        {
        if(pIp[iPos]==0)break;
        cName[iPos]=(char)pIp[iPos];
        }

    cName[iPos]=0;


return IpAddressExA(pAddr,cName,pLen);
}
#endif

//*****************************************************************************
//*
//*     IpGetHostname
//*
//*****************************************************************************
//  Gets the hostname of this station.
//  Returns a pointer to the hostname
SUAPI const char *IpGetHostname()
{

    #if !defined(RTTARGET) && !defined(LWIP)

    if(!cHostName[0])
        {
        InitSocket();
        gethostname(cHostName,sizeof(cHostName));
        }

    #endif

return cHostName;
}

//*****************************************************************************
//*
//*     IpSetHostname
//*
//*****************************************************************************
//  Sets the hostname for this station
//  Returns the length of the hostname
SUAPI int IpSetHostname(const char *pName)
{
int     iPos;


    for(iPos=0;iPos<(int)sizeof(cHostName)-1;iPos++)
        {
        if(!pName[iPos])break;
        cHostName[iPos] = pName[iPos];
        }

    cHostName[iPos] = 0;


return iPos;
}

//*****************************************************************************
//*
//*     IpIsV6Address
//*
//*****************************************************************************
//  Check a name if it is an IP V6 ip address, i.e. :
//              "fd9e:21a7:a92c:2323::1"
//              "fc00::/7"
//              "2001:db8::2"
//              "[2001:db8::2]"
//              "[::ffff:192.168.32.1]"
//              "::1"
//              "::"
//  pName   : is the pointer to the name
//  Returns >=1 if it an IP V6 ip address, or 0 in the other case
//                  0 = is a hostname or an IPv4 address
//                  1 = a valid IPv6 address  "1:2:3:4:5:6:7:8"
//                  2 = any address "::"
//                  3 = loopback address "::1"
//                >=7 = offset from sup IP "::ffff:192.168.0.1"
SUAPI int IpIsV6AddressA(const char *pName)
{
char            cSign;
const char     *pStart;
unsigned short  aWords[8+5];
unsigned        uValue;
int             iDouble;
int             iBraket;
int             iPoint;
int             iLast;
int             iSegs;
int             iSub;
int             iLen;
int             iRet;




    pStart  = pName;
    uValue  = 0;
    iBraket = 0;
    iDouble = 0;
    iPoint  = 0;
    iSegs   = 0;
    iLast   = 1;
    iLen    = 0;
    iSub    = 0;


    if(*pName=='[')
        {
        iBraket=1;
        pName++;
        }


    for(;;pName++)
        {
           cSign = *pName;
        if(cSign==0)break;
        if(cSign=='.')                                      // special IPv4 "::FFFF:192.168.0.1"
            {
            if(uValue>0x255)return 0;

            if(iPoint==0)                                   // first IPv4 part
                {
                if(iSegs<1)return 0;
                if(iSegs>6)return 0;
                if(iSegs<6 && !iDouble)return 0;
                if(uValue%16>9)return 0;

                iSegs--;

                if(aWords[iSegs]!=0xFFFF)return 0;

                for(iSegs--;iSegs>=0;iSegs--)
                    {
                    if(aWords[iSegs])return 0;
                    }

                iPoint = iLast;
                iSub   = 1;
                }
            else if(iSub>=3)                                // only 4 sub parts
                {
                return 0;
                }
            else{
                iSub++;
                }


            uValue = 0;
            iLen   = 0;

            continue;
            }

        if(cSign==':')
            {
            aWords[iSegs] = (unsigned short)uValue;
            uValue        = 0;

               iSegs++;
            if(iSegs>=8 || iPoint)return 0;

            iLen = 0;

            if(pName[1]==':')                               // found "::"
                {
                if(iDouble      )return 0;
                if(iSegs>=7     )return 0;
                if(pName[2]==':')return 0;
                if(pName[2]==']' && !iBraket)return 0;

                aWords[iSegs] = 0;
                pName   += 1;
                iSegs   += 1;
                iDouble  = iSegs;

                continue;
                }

            if(pName[1]== 0 )return 0;                      // "1234:"  is not allowed
            if(pName[1]==']')return 0;                      // "[234:]" is not allowed

            iLast = pName-pStart+1;

            continue;
            }


        if(cSign<='9' && cSign>='0')
            {
            uValue <<= 4;
            uValue  += cSign-'0';
            iLen    += 1;
            }
        else if(cSign==']')                                 // check end of text
            {
            if(iBraket)break;
            return 0;
            }
        else if(iPoint)
            {
            return 1;
            }
        else if(cSign>='a' && cSign<='f')
            {
            uValue <<= 4;
            uValue  += cSign-'a'+10;
            iLen    += 1;
            }
        else if(cSign>='A' && cSign<='F')
            {
            uValue <<= 4;
            uValue  += cSign-'A'+10;
            iLen    += 1;
            }
        else{
            return 0;
            }

        if(iLen>4)return 0;
        }



    if(iPoint)                                              // special IPv4 "::FFFF:192.168.0.1"
        {
        if(iSub==3 && uValue<=0x255)return iPoint;
        }
    else{
        if(iSegs==7 || iDouble)
            {
            if(uValue==0)                                   // is it any address "::"
                {
                iRet = 2;
                }
            else if(uValue==1)                              // is it loopback "::1"
                {
                iRet = 3;
                }
            else{
                return 1;
                }

            for(iSegs--;iSegs>=0;iSegs--)                   // check for zerros
                {
                if(aWords[iSegs])return 1;
                }

            return iRet;
            }
        }




return 0;
}

//*****************************************************************************
//*
//*     IpIsV6AddressA
//*
//*****************************************************************************
//  Check a name if it is an IP V6 ip address, i.e. :
//              "fd9e:21a7:a92c:2323::1"
//              "fc00::/7"
//              "2001:db8::2"
//              "[2001:db8::2]"
//              "[::ffff:192.168.32.1]"
//              "::1"
//              "::"
//  pName   : is the pointer to the name
//  Returns >=1 if it an IP V6 ip address, or 0 in the other case
//                  0 = is a hostname or an IPv4 address
//                  1 = a valid IPv6 address  "1:2:3:4:5:6:7:8"
//                  2 = any address "::"
//                  3 = loopback address "::1"
//                >=7 = offset from sup IP "::ffff:192.168.0.1"
int IpIsV6AddressW(const wchar_t *pName)
{
char    cName[64];
int     iPos;


    for(iPos=0;iPos<63;iPos++)
        {
        if(pName[iPos]==0)break;
        cName[iPos]=(char)pName[iPos];
        }

    cName[iPos]=0;


return IpIsV6AddressA(cName);
}

//*****************************************************************************
//*
//*     IpHostname
//*
//*****************************************************************************
//  Converts a text to an IP Address
//  pAddr       : Here is the IP address saved
//  pHostname   : Is the  i.e. "127.0.0.1" "www.sbox.tugraz.at" "ax-progr-22"
//                NULL is the own host name
//  Returns 0 if an error occurs or 1 for a IPV4 address or 2 for IPV6 address.
SUAPI int IpHostnameA(IpAddr *pAddr,const char *pHostname)
{
char                 cMyName[128];
char               **pAddresses;
struct hostent      *pHost;
struct addrinfo_ex  *pData;
struct addrinfo_ex   sInfo;
struct addrinfo_ex  *pPtr;
struct sockaddr_in  *pIp;
int                  iRet;
int                  iLen;



    InitSocket();

    pData = 0;

    if(pHostname)
        {                                       // don't resolve IPV6 addresses if only IPV4 pressend
        #if SU_USE_IPV6
        if(iIpV6Allowed==0)
        #endif
            {
               iRet = IpIsV6AddressA(pHostname);
            if(iRet==0)                         // it isn't an IPv6 addess
                {
                }
            else if(iRet==2)                    // any address "::" used as "0.0.0.0"
                {
                SET_V4_ADDR(pAddr,0);
                return 1;
                }
            else if(iRet==3)                    // loopback "::1" used as "127.0.0.1"
                {
                SET_V4_ADDR(pAddr,LOOPBACK);
                return 1;
                }
            else if(iRet<7)                     // can't convert IPv6 to an IPv4 address
                {
                SET_V4_ADDR(pAddr,INADDR_NONE);
                return 0;
                }
            else if(*pHostname=='[')            // sub ip "[::FFFF:192.168.0.1]"
                {
                pHostname += iRet;

                for(iLen=0;;iLen++)
                    {
                    if(pHostname[iLen]==']')break;
                    cMyName[iLen]=pHostname[iLen];
                    }

                cMyName[iLen]=0;
                pHostname = cMyName;
                }
            else{                               // sub ip "::FFFF:192.168.0.1"
                pHostname += iRet;
                }
            }


        memset(&sInfo,0,sizeof(sInfo));

        #ifdef LINUX

        if(IpIsV6Address(pHostname))            // linux means "fe80::318e..." is an alphabethic hostname
            {
            sInfo.ai_flags |= AI_NUMERICHOST;
            sInfo.ai_family = AF_INET6;
            }

        #endif


           iRet = getaddrinfo_ex(pHostname,0,&sInfo,&pData);
        if(iRet==0)
            {
            pPtr = pData;

            while(pPtr)
                {
                pIp = (struct sockaddr_in*)(pPtr->ai_addr);

                if(iRet!=1 && pIp->sin_family==AF_INET)
                    {
                    SET_V4_ADDR(pAddr,*(unsigned*)&pIp->sin_addr);
                    iRet = 1;
                    break;
                    }

                #if SU_USE_IPV6
                else if(iRet==0 && pIp->sin_family==AF_INET6 && iIpV6Allowed)
                    {
                    struct sockaddr_in6_ *pIp6=(struct sockaddr_in6_*)pIp;
                    memcpy(pAddr->aAddr,pIp6->sin6_addr.s6_addr,16);
                    iRet = 2;
                    }
                #endif

                pPtr = pPtr->ai_next;
                }


            freeaddrinfo_ex(pData);

            if(iRet)return iRet;
            }
        }


//****************** old version with gethostbyname ***************************


    if(!pHostname)                              // use the own hostname
        {
        #if defined(RTTARGET) || defined(LWIP)

        if(cHostName[0]==0)
            {
            SET_V4_ADDR(pAddr,INADDR_NONE);
            return 0;
            }

        #else

        if(gethostname(cMyName,sizeof(cMyName))== SOCKET_ERROR)
            {
            SET_V4_ADDR(pAddr,INADDR_NONE);
            return 0;
            }

        #endif

        pHostname = cMyName;
        }

        pHost=gethostbyname((char*)pHostname);
    if(!pHost)
        {
        SET_V4_ADDR(pAddr,INADDR_NONE);
        return 0;
        }

    pAddresses=pHost->h_addr_list;

    if(!pAddresses[0])
        {
        SET_V4_ADDR(pAddr,INADDR_NONE);
        return 0;
        }


       SET_V4_ADDR(pAddr,*(unsigned*)pAddresses[0]);
    if(GET_V4_ADDR(pAddr)==INADDR_NONE)return 0;



return 1;
}

//*****************************************************************************
//*
//*     IpAddressW
//*
//*****************************************************************************
#if defined(WIN32)
int IpHostnameW(IpAddr *pAddr,const wchar_t *pHostname)
{
char    cHost[64];
int     iPos;


    if(pHostname)
        {
        for(iPos=0;iPos<63;iPos++)
            {
            if(pHostname[iPos]==0)break;
            cHost[iPos]=(char)pHostname[iPos];
            }

        cHost[iPos]=0;
        }


return IpHostnameA(pAddr,(pHostname)? cHost:0);
}
#endif



//*****************************************************************************
//*
//*     UdpHandleAlloc
//*
//*****************************************************************************/
//  Creates a new UDP handel
//  Socket      : Socket number for the handle
//  pAddr       : pointer to the "sockaddr_in" data
//  iFlags      : several flags for the socket
//                  UDP_BLOCKING    0x00000000 = activate the blocking mode
//                  UDP_NOBLOCKING  0x00000001 = activate none-blocking mode
//                  UDP_BCASTSEND   0x00000002 = allow broadcast sending (255.255.255.255)
//                  UDP_BCASTRECV   0x00000004 = allow broadcast receiving
//                  UDP_REUSEADDR   0x00000008 = enable reuse of the address
//                  UDP_CLOSEDOWN   0x00000010 = colse socket at UdpShutdown(...)
//  Returns the pointer to the UDP handle
static UdpHandle UdpHandleAlloc(SOCKET iSocket,void *pAddr,int iFlags)
{
UdpData *pData;


        pData=(UdpData*)MallocHandle(sizeof(UdpData));
    if(!pData)
        {
        closesocket(iSocket);
        return 0;
        }



    pData->iSocket      = iSocket;
    pData->ucSockFlags  = (unsigned char)( iFlags );
    pData->ucConnFlags  = (unsigned char)((iFlags&UDP_CLOSEDOWN)? FLAG_DOWN|FLAG_WRITE:FLAG_WRITE);
    pData->hDummy       = TAG_UDP;

    #ifdef SU_WINDOWS
    pData->hEvent = WSACreateEvent();
    WSAEventSelect(pData->iSocket,pData->hEvent,FD_READ|FD_WRITE);
    #endif

    #ifdef RTTARGET
    pData->hSemaphor = RTKCreateSemaphore(ST_COUNTING,1,NULL);
    #endif

    #if SU_USE_IPV6

    memcpy(&pData->sAddr,pAddr,sizeof(pData->sAddr6));

    if(pData->sAddr.sin_family==AF_INET)
        {
        pData->sIpAddr.aAddr[0]=SU_IPV4_ID;
        pData->sIpAddr.aAddr[1]=pData->sAddr.sin_addr.s_addr;
        pData->usPort   = htons(pData->sAddr.sin_port);
        }
    else{
        memcpy(pData->sIpAddr.aAddr,pData->sAddr6.sin6_addr.s6_addr,16);
        pData->usPort = htons(pData->sAddr6.sin6_port);
        }

    #else

    memcpy(&pData->sAddr,pAddr,sizeof(pData->sAddr));

    pData->sIpAddr.aAddr[0] = pData->sAddr.sin_addr.s_addr;
    pData->usPort     = htons(pData->sAddr.sin_port);

    #endif


return (UdpHandle)pData;
}


//*****************************************************************************
//*
//*     UdpHandleFree
//*
//*****************************************************************************/
//  Releases an UDP handle
static inline void UdpHandleFree(UdpHandle hHandle)
{

    if(!hHandle)return;

    #ifdef SU_WINDOWS

    if(((UdpData*)hHandle)->hEvent)
        {
        WSACloseEvent(((UdpData*)hHandle)->hEvent);
        ((UdpData*)hHandle)->hEvent = 0;
        }

    #endif

    #ifdef RTTARGET

    RTKDeleteSemaphore(&((UdpData*)hHandle)->hSemaphor);

    #endif

    FreeHandle(hHandle);
}


//*****************************************************************************
//*
//*     UdpOpen
//*
//*****************************************************************************/
//  Creates a UDP socket
//  pAddr       : Is the IP adress for the socked
//  uPort       : UDP port number (0=assign an unique port numper)
//  iFlags      : Several falgs for the UDP socket
//                  UDP_BLOCKING    0x00000000 = activate the blocking mode
//                  UDP_NOBLOCKING  0x00000001 = activate none-blocking mode
//                  UDP_BCASTSEND   0x00000002 = allow broadcast sending (255.255.255.255)
//                  UDP_BCASTRECV   0x00000004 = allow broadcast receiving
//                  UDP_REUSEADDR   0x00000008 = enable reuse of the address
//                  UDP_CLOSEDOWN   0x00000010 = colse socket at UdpShutdown(...)
//  Retrns 0 at an error, or a UDP handle
SUAPI UdpHandle UdpOpen(IpAddr *pAddr,unsigned short uPort,int iFlags)
{
void                   *pData;
struct sockaddr_in      sData;
#if SU_USE_IPV6
struct sockaddr_in6_    sData6;
#endif
SOCKET                  iSock;
IpAddr                  iTemp;
unsigned                aAddr;
unsigned                uOrig=0;
int                     iErr;
unsigned long           lNoneBlocking=1;



    InitSocket();

    iSock     = CreateServerSocket(&pAddr,&iTemp,1);


    #if SU_USE_IPV6

    if(pAddr->aAddr[0]!=SU_IPV4_ID)
        {                                               // open socket
        if(iSock==INVALID_SOCKET)iSock=socket(AF_INET6,SOCK_DGRAM,IPPROTO_UDP);
        if(iSock==INVALID_SOCKET)
            {
            return 0;
            }

        if(iFlags&UDP_REUSEADDR)                        // enable reuse of the address
            {
            int iVal;

            iVal=1;

            if(setsockopt(iSock,SOL_SOCKET,SO_REUSEADDR,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
                {
                closesocket(iSock);
                return 0;
                }
            }

        memset(&sData6,0,sizeof(sData6));               //  build local address

        sData6.sin6_family      =  AF_INET6;
        sData6.sin6_port        =  htons(uPort);

        if(iFlags&UDP_BCASTSEND)
                memcpy(sData6.sin6_addr.s6_addr,MULTICAST_V6,16);
        else    memcpy(sData6.sin6_addr.s6_addr,pAddr->aAddr,16);

        iErr  = bind(iSock,(const struct sockaddr*)&sData6,sizeof(sData6));
        pData = &sData6;
        }
    else{
    #endif                                              // open socket
        if(iSock==INVALID_SOCKET)iSock=socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
        if(iSock==INVALID_SOCKET)
            {
            return 0;
            }

        memset(&sData,0,sizeof(sData));                 //  build local address

        aAddr = GET_V4_ADDR(pAddr);
        uOrig = aAddr;

        #ifdef LINUX

        if((iFlags&UDP_BCASTRECV) && !IsIpAny(pAddr))   // the fucking LINUX can't handle receive broadcasts correct
        if(BindToInterface(iSock,aAddr))
            {
            iFlags &= ~UDP_REUSEADDR;                   // bind to interface
            }

        #endif


        if(iFlags&UDP_REUSEADDR)                        // enable reuse of the address
            {
            int iVal;

            iVal=1;

            if(setsockopt(iSock,SOL_SOCKET,SO_REUSEADDR,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
                {
                closesocket(iSock);
                return 0;
                }
            }


        #if defined(RTTARGET) && RTT32_VER<523
        if(iFlags&UDP_BCASTRECV)aAddr = INADDR_ANY;     // RTOS don't receive broadcasts on common ports
        #endif

        #ifdef LINUX
        if(iFlags&UDP_BCASTRECV)aAddr = INADDR_ANY;     // the fucking LINUX can't handle SO_BROADCAST correct
        #endif

        sData.sin_family      =  AF_INET;
        sData.sin_addr.s_addr =  aAddr;
        sData.sin_port        =  htons(uPort);

        iErr  = bind(iSock,(const struct sockaddr*)&sData,sizeof(sData));
        pData = &sData;
    #if SU_USE_IPV6
        }
    #endif


    if(iErr)
        {
        closesocket(iSock);
        return 0;
        }

    if(iFlags&TCP_NOBLOCKING)
        {
        lNoneBlocking = 1;
        if(ioctlsocket(iSock,FIONBIO,&lNoneBlocking))
            {
            closesocket(iSock);
            return 0;
            }
        }

    #ifdef LWIP

    if(iFlags&(UDP_BCASTSEND|UDP_BCASTRECV))
        {
        int iVal=1;

        if(setsockopt(iSock,SOL_SOCKET,SO_BROADCAST,(char*)&iVal,sizeof(iVal)) == -1)
            {
            closesocket(iSock);
            return 0;
            }
        }

    #elif !defined(RTTARGET) || RTT32_VER>=523

    if(iFlags&UDP_BCASTSEND)
        {
        int iVal=1;

        if(setsockopt(iSock,SOL_SOCKET,SO_BROADCAST,(char*)&iVal,sizeof(iVal)) == -1)
            {
            closesocket(iSock);
            return 0;
            }
        }

    #endif

    if(uOrig)                                           // change back to original ip
        {
        sData.sin_addr.s_addr = uOrig;
        }



return UdpHandleAlloc(iSock,pData,iFlags);
}


//*****************************************************************************
//*
//*     UdpBlock
//*
//*****************************************************************************/
//  Changes the blocking mode of a UDP socket
//  hHandle     : UDP handle of the socket
//  iOn         : 0=none-blocking  1=blocking
//  Returns 0 if an error occurs or 1 if all ok
SUAPI int UdpBlock(UdpHandle hHandle,int iOn)
{
UdpData        *pData=(UdpData*)hHandle;


    if(!pData)return 0;

    if(!iOn && !(pData->ucSockFlags&UDP_NOBLOCKING))
        {
        #ifndef SU_WINDOWS                              // windows has always noneblocking sockets
        unsigned long  lNoneBlocking = 1;
        if(ioctlsocket(pData->iSocket,FIONBIO,&lNoneBlocking))return 0;
        #endif

        pData->ucSockFlags |=  UDP_NOBLOCKING;
        }

    if(iOn && (pData->ucSockFlags&UDP_NOBLOCKING))
        {
        #ifndef SU_WINDOWS                              // windows has always noneblocking sockets
        unsigned long  lNoneBlocking = 0;
        if(ioctlsocket(pData->iSocket,FIONBIO,&lNoneBlocking))return 0;
        #endif

        pData->ucSockFlags &= ~UDP_NOBLOCKING;
        }


return 1;
}

//*****************************************************************************
//*
//*     UdpCloseDown
//*
//*****************************************************************************/
//  Changes the close down mode of a UDP socket
//  hHandle     : UDP handle of the socket
//  iOn         : 0=not close at shutdown 1=close at shutdown
//  Returns 0 if an error occurs or 1 if all ok
SUAPI int UdpCloseDown(UdpHandle hHandle,int iOn)
{
UdpData        *pData=(UdpData*)hHandle;


    if(!pData)return 0;

    if(iOn)pData->ucConnFlags|= UDP_CLOSEDOWN;
    else   pData->ucConnFlags&=~UDP_CLOSEDOWN;


return 1;
}

//*****************************************************************************
//*
//*     UdpSetBuffers
//*
//*****************************************************************************/
//  Changes the send and receive buffer sizes of a socket
//  hHandle     : TCP handle of the socket
//  iRecvSize   : Size for the receive buffer (0=none-change)
//  iSendSize   : Size for the send    buffer (0=none-change)
//  Returns     Bit0 :  0=receive buffer not change  1=changed
//              Bit1 :  0=send    buffer not change  1=changed
SUAPI int UdpSetBuffers(UdpHandle hHandle,int iRecvSize,int iSendSize)
{
UdpData        *pData=(UdpData*)hHandle;
int             iRet=0;


    if(!pData)return 0;

    #ifndef RTTARGET

    if(iRecvSize && setsockopt(pData->iSocket,SOL_SOCKET,SO_RCVBUF,(char*)&iRecvSize,sizeof(iRecvSize)) != -1)
        {
        iRet |=1;
        }

    if(iSendSize && setsockopt(pData->iSocket,SOL_SOCKET,SO_SNDBUF,(char*)&iSendSize,sizeof(iSendSize)) != -1)
        {
        iRet |=2;
        }

    #endif


return iRet;
}


//*****************************************************************************
//*
//*     UdpShutdown
//*
//*****************************************************************************/
//  Disconnects a UDP connection. After this call no data will send or receive
//  on the socket
//  hHandle     : Handle of the UDP socket
//  Returns 0 if an error occurs or 1 if all ok
SUAPI int UdpShutdown(UdpHandle hHandle)
{
UdpData    *pData=(UdpData*)hHandle;
int         iSocket;


    if(!hHandle)return 0;
    if(shutdown(pData->iSocket,2))return 0;

    if(pData->ucConnFlags&UDP_CLOSEDOWN)
        {
        iSocket            = pData->iSocket;
        pData->iSocket     = -1;
        closesocket(iSocket);
        }


return 1;
}

//*****************************************************************************
//*
//*     UdpClose
//*
//*****************************************************************************/
//  Closes a UDP socket
//  hHandle     : Handle of the UDP socket
//  Returns 0 if an error occurs or 1 if all ok
SUAPI int UdpClose(UdpHandle hHandle)
{
UdpData   *pData=(UdpData*)hHandle;


    if(!hHandle || !IS_TAG_UDP(pData->hDummy))
        {
        return 0;
        }

    shutdown   (pData->iSocket,2);
    closesocket(pData->iSocket  );
    pData->hDummy = 0;
    UdpHandleFree(hHandle);


return 1;
}

//*****************************************************************************
//*
//*     UdpPut
//*
//*****************************************************************************/
//  Sends a packet to a destination address, to the same port
//  pAddr : IP-Address of the destination
//  pMsg     : Pointer to the packet
//  iLen     : Size of the packet
//  Returns the count of the sended iBytes, or zero at an error
SUAPI int UdpPut(UdpHandle hHandle,IpAddr *pAddr,const void *pMsg,int iLen)
{
UdpData            *pData=(UdpData*)hHandle;
struct sockaddr_in  sInput;
unsigned short      usPort;
int                 iBytes;
int                 iErr=0;


    if(!hHandle)return 0;

    usPort = pData->sAddr.sin_port;

    #if SU_USE_IPV6

    if(pData->sAddr.sin_family==AF_INET6)
        {
        usPort = pData->sAddr6.sin6_port;
        }

    if(pAddr->aAddr[0]!=SU_IPV4_ID)
        {
        struct sockaddr_in6_   sInput6;

RetryA: sInput6.sin6_family =  AF_INET6;
        sInput6.sin6_port   =  usPort;
        memcpy(&sInput6.sin6_addr,pAddr->aAddr,16);

        #ifdef RTTARGET
        RTKWait(pData->hSemaphor);
        #endif

        iBytes = sendto(pData->iSocket,(char*)pMsg,iLen,SEND_FLAGS,(const struct sockaddr*)&sInput6,sizeof(sInput6));

        #ifdef RTTARGET
        RTKSignal(pData->hSemaphor);
        #endif

        if(iBytes<iLen)
            {
            pData->ucConnFlags &= ~FLAG_WRITE;
            if((pData->ucSockFlags&UDP_NOBLOCKING) || iBytes>0 || iErr>4)return 0;
            if(SocketWaitWr(250,PARAM_WAIT(pData))<=0)return 0;
            iErr++;
            goto RetryA;
            }

        return iBytes;
        }

    #endif

RetryB:

    sInput.sin_family       =  AF_INET;
    sInput.sin_addr.s_addr  =  GET_V4_ADDR(pAddr);
    sInput.sin_port         =  usPort;

    #ifdef RTTARGET
    RTKWait(pData->hSemaphor);
    #endif

    iBytes = sendto(pData->iSocket,(char*)pMsg,iLen,SEND_FLAGS,(const struct sockaddr*)&sInput,sizeof(sInput));

    #ifdef RTTARGET
    RTKSignal(pData->hSemaphor);
    #endif

    if(iBytes<iLen)
        {
        pData->ucConnFlags &= ~FLAG_WRITE;
        if((pData->ucSockFlags&UDP_NOBLOCKING) || iBytes>0 || iErr>4)return 0;
        if(SocketWaitWr(250,PARAM_WAIT(pData))<=0)return 0;
        iErr++;
        goto RetryB;
        }


return iBytes;
}

//*****************************************************************************
//*
//*     UdpPutTo
//*
//*****************************************************************************/
//  Sends a packet to a destination address to a other port
//  pAddr   : IP-Address of the destination
//  pMsg    : Pointer to the packet
//  iLen    : Size of the packet
//  uPort   : Port number of the destination
//  Returns the count of the sended bytes, or zero at an error
SUAPI int UdpPutTo(UdpHandle hHandle,IpAddr *pAddr,const void *pMsg,int iLen,unsigned short uPort)
{
UdpData            *pData=(UdpData*)hHandle;
struct sockaddr_in  sInput;
int                 iBytes;
int                 iErr=0;


    if(!hHandle)return 0;

    #if SU_USE_IPV6

    if(pAddr->aAddr[0]!=SU_IPV4_ID)
        {
        struct sockaddr_in6_    sInput6;

RetryA: sInput6.sin6_family =  AF_INET6;
        sInput6.sin6_port   =  htons(uPort);
        memcpy(&sInput6.sin6_addr,pAddr->aAddr,16);

        #ifdef RTTARGET
        RTKWait(pData->hSemaphor);
        #endif

        iBytes = sendto(pData->iSocket,(char*)pMsg,iLen,SEND_FLAGS,(const struct sockaddr*)&sInput6,sizeof(sInput6));

        #ifdef RTTARGET
        RTKSignal(pData->hSemaphor);
        #endif

        if(iBytes<iLen)
            {
            pData->ucConnFlags &= ~FLAG_WRITE;
            if((pData->ucSockFlags&UDP_NOBLOCKING) || iBytes>0 || iErr>4)return 0;
            if(SocketWaitWr(250,PARAM_WAIT(pData))<=0)return 0;
            iErr++;
            goto RetryA;
            }

        return iBytes;
        }

    #endif

RetryB:

    sInput.sin_family       =  AF_INET;
    sInput.sin_addr.s_addr  =  GET_V4_ADDR(pAddr);
    sInput.sin_port         =  htons(uPort);

    #ifdef RTTARGET
    RTKWait(pData->hSemaphor);
    #endif

    iBytes = sendto(pData->iSocket,(char*)pMsg,iLen,SEND_FLAGS,(const struct sockaddr*)&sInput,sizeof(sInput));

    #ifdef RTTARGET
    RTKSignal(pData->hSemaphor);
    #endif

    if(iBytes<iLen)
        {
        pData->ucConnFlags &= ~FLAG_WRITE;
        if((pData->ucSockFlags&UDP_NOBLOCKING) || iBytes>0 || iErr>4)return 0;
        if(SocketWaitWr(250,PARAM_WAIT(pData))<=0)return 0;
        iErr++;
        goto RetryB;
        }


return iBytes;
}

//*****************************************************************************
//*
//*     UdpGet
//*
//*****************************************************************************/
//  Reseives a packet
//  Form        : UDP handle from the socket
//  pRecv       : If this pointer is not zero,the ip address of the
//                sender will stored there.
//  pMsg        : Pointer to the packet buffer
//  iLen        : Size of the packet buffer
//  Returns the count of the sended bytes, or zero at an error
SUAPI int UdpGet(UdpHandle hFrom,IpAddr *pRecv,void *pMsg,int iLen,unsigned short *pPort)
{
UdpData                *pData=(UdpData*)hFrom;
#if SU_USE_IPV6
struct sockaddr_in6_    sInput;
#else
struct sockaddr_in      sInput;
#endif
int                     iBytes;
int                     iSize;


    if(!hFrom)                                          // check handle
        {
        if(pRecv)SET_V4_ADDR(pRecv,0);
        if(pPort)*pPort=0;
        return -1;
        }


    #if SU_USE_IPV6
    sInput.sin6_family = pData->sAddr.sin_family;
    #endif

    #ifdef SU_WINDOWS

    if(!(pData->ucSockFlags&UDP_NOBLOCKING))            // windows has always none blocking sockets
        {
        int iWait = ((RECV_TIMEOUT)? RECV_TIMEOUT*1000:30000);
        SocketWaitRd(iWait,PARAM_WAIT(pData));
        }

    #endif

       iSize  = sizeof(sInput);
       iBytes = recvfrom(pData->iSocket,(char*)pMsg,iLen,RECV_FLAGS,(struct sockaddr*)&sInput,(LEN_TYPE)&iSize);
    if(iBytes<=0)
        {
        if(pRecv)SET_V4_ADDR(pRecv,0);
        if(pPort)*pPort=0;
        return iBytes;
        }

    #if SU_USE_IPV6
    if(sInput.sin6_family==AF_INET)
        {
        if(pRecv)SET_V4_ADDR(pRecv,((struct sockaddr_in*)&sInput)->sin_addr.s_addr);
        if(pPort)*pPort = htons   (((struct sockaddr_in*)&sInput)->sin_port);
        }
    else{
        if(pRecv)memcpy(pRecv->aAddr,sInput.sin6_addr.s6_addr,16);
        if(pPort)*pPort=htons(sInput.sin6_port);
        }
    #else
    if(pRecv)SET_V4_ADDR(pRecv,sInput.sin_addr.s_addr);
    if(pPort)*pPort=htons(sInput.sin_port);
    #endif


return iBytes;
}

//*****************************************************************************
//*
//*     UdpGetWait
//*
//*****************************************************************************/
//  Reseives a packet with a timeout. The function returns only
//  if a packet receives, or the timeout is expired.
//  hForm       : UDP handle from the socket
//  pRecv       : If this pointer is not zero,the ip address of the
//                sender will stored there.
//  pMsg        : Pointer to the packet buffer
//  iLen        : Size of the packet buffer
//  iTimeout    : Time in millisecounds for the timeout
//  pPort       : If this pointer is not zero,the receifed port address
//  Returns the count of the sended iBytes, or zero at an error
SUAPI int UdpGetWait(UdpHandle hFrom,IpAddr *pRecv,void *pMsg,int iLen,int iTimeout,unsigned short *pPort)
{
UdpData                *pData=(UdpData*)hFrom;
#if SU_USE_IPV6
struct sockaddr_in6_    sInput;
#else
struct sockaddr_in      sInput;
#endif
int                     iBytes;
int                     iCount;
int                     iSize;


    if(!hFrom)                                          // check handle
        {
        if(pRecv)SET_V4_ADDR(pRecv,0);
        if(pPort)*pPort=0;
        return -1;
        }

    if(iTimeout<0)                                      // check timeout value
        {
        if(pRecv)SET_V4_ADDR(pRecv,0);
        if(pPort)*pPort=0;
        return 0;
        }


       iCount=SocketWaitRd(iTimeout,PARAM_WAIT(pData)); // wait until packet received
    if(iCount<=0)
        {
        if(pRecv)SET_V4_ADDR(pRecv,0);
        if(pPort)*pPort=0;
        return iCount;
        }

    #if SU_USE_IPV6
    sInput.sin6_family = pData->sAddr.sin_family;
    #endif

       iSize  = sizeof(sInput);
       iBytes = recvfrom(pData->iSocket,(char*)pMsg,iLen,RECV_FLAGS,(struct sockaddr*)&sInput,(LEN_TYPE)&iSize);
    if(iBytes<=0)
        {
        if(pRecv)SET_V4_ADDR(pRecv,0);
        if(pPort)*pPort=0;
        return iBytes;
        }


    #if SU_USE_IPV6
    if(sInput.sin6_family==AF_INET)
        {
        if(pRecv)SET_V4_ADDR(pRecv,((struct sockaddr_in*)&sInput)->sin_addr.s_addr);
        if(pPort)*pPort = htons   (((struct sockaddr_in*)&sInput)->sin_port);
        }
    else{
        if(pRecv)memcpy(pRecv->aAddr,sInput.sin6_addr.s6_addr,16);
        if(pPort)*pPort=htons(sInput.sin6_port);
        }
    #else
    if(pRecv)SET_V4_ADDR(pRecv,sInput.sin_addr.s_addr);
    if(pPort)*pPort=htons(sInput.sin_port);
    #endif



return iBytes;
}

//*****************************************************************************
//*
//*     UdpWait
//*
//*****************************************************************************/
//  Waits until the next data from a UDP socket was received.
//  hHandle     : Handle of the UDP socket
//  iTimeout    : Timeout in millisecounds
//  Returns -1 if an error occurs.
//  Returns  0 on timeouts.
//  Returns  the count of data in the the input buffer.
SUAPI int UdpWait(UdpHandle hHandle,int iTimeout)
{
UdpData        *pData=(UdpData*)hHandle;
unsigned long   uReady;



    if(hHandle==0)return -1;
    if(iTimeout<0)
        {
        return 0;
        }

    if(SocketWaitRd(iTimeout,PARAM_WAIT(pData))<=0)     // wait until data received
        {
        return 0;
        }

    if(ioctlsocket(pData->iSocket,FIONREAD,&uReady))    // get byte count in queue
        {
        return -1;
        }


return uReady;
}


//*****************************************************************************
//*
//*     UdpGetCount
//*
//*****************************************************************************/
//  Returns the size of the next packet in the receive buffer of the socket
//  hHandle : UDP handle from the socket
//  Return the size or 0 if no packet is in the buffer, or -1 on an error
SUAPI int UdpGetCount(UdpHandle hHandle)
{
UdpData        *pData=(UdpData*)hHandle;
unsigned long   uDataReady;


    if(!hHandle)return 0;
    if(ioctlsocket(pData->iSocket,FIONREAD,&uDataReady))return -1;


return uDataReady;
}


//*****************************************************************************
//*
//*     UdpGetLocal
//*
//*****************************************************************************
//  Gets the IP address of the local side of the handle
//  hHandle     : UDP handle of the socket
//  pAddr       : Poiter to the adress memory (0 means not get)
//  pPort       : Pointer fpr the port number (0 means not get)
//  Returns 0 on errors or 1 if ok
SUAPI int UdpGetLocal(UdpHandle hHandle,IpAddr *pAddr,unsigned short *pPort)
{
UdpData                *pData=(UdpData*)hHandle;
#if SU_USE_IPV6
struct sockaddr_in6_    sAddr;
#else
struct sockaddr_in      sAddr;
#endif
int                     iLen;



    iLen = sizeof(sAddr);

    if(!hHandle || getsockname(pData->iSocket,(struct sockaddr*)&sAddr,(LEN_TYPE)&iLen))
        {
        #if SU_USE_IPV6

        if(pAddr)
            {
            pAddr->aAddr[0]=SU_IPV4_ID;
            pAddr->aAddr[1]=0;
            }

        if(pPort)*pPort=0;

        #else

        if(pAddr)SET_V4_ADDR(pAddr,0);
        if(pPort)*pPort=0;

        #endif

        return 0;
        }



    #if SU_USE_IPV6

    if(sAddr.sin6_family==AF_INET)
        {
        if(pAddr)SET_V4_ADDR(pAddr,((struct sockaddr_in*)&sAddr)->sin_addr.s_addr);
        if(pPort)*pPort = htons   (((struct sockaddr_in*)&sAddr)->sin_port);
        }
    else{
        if(pAddr)memcpy(pAddr->aAddr,sAddr.sin6_addr.s6_addr,16);
        if(pPort)*pPort=htons(sAddr.sin6_port);
        }

    #else

    if(pAddr)SET_V4_ADDR(pAddr,sAddr.sin_addr.s_addr);
    if(pPort)*pPort=htons(sAddr.sin_port);

    #endif



return 1;
}

//*****************************************************************************
//*
//*     TcpHandleAlloc
//*
//*****************************************************************************/
//  Creates a new TCP handel
//  iSocket     : socket number for the pData
//  pAddr       : pointer to the "sockaddr_in" data
//  iFlags      : several flags for the socket
//                  TCP_NOBLOCKING  0x00000001  = sets the socket in the no blocking mode
//                  TCP_NODELAYACK  0x00000002  = disable nagle algorithm
//                  TCP_KEEP_ALIVE  0x00000004  = enable sending of keepalive packets
//                  TCP_REUSEADDR   0x00000008  = allows to reuse an used tcp port
//                  TCP_CLOSEDOWN   0x00000010  = close socket at TcpShutdown(...)
//                  TCP_DATABROKEN  0x00000020  = no broken signal if data received
//  Returns the pointer to the handle data
static TcpHandle TcpHandleAlloc(SOCKET iSocket,void *pAddr,int iServer,int iFlags)
{
TcpData *pData;


        pData=(TcpData*)MallocHandle(sizeof(TcpData));
    if(!pData)
        {
        closesocket(iSocket);
        return 0;
        }


    pData->iSocket      =  iSocket;
    pData->ucSockFlags  = (unsigned char)( iFlags );
    pData->ucConnFlags  = (unsigned char)((iFlags&TCP_CLOSEDOWN)? FLAG_DOWN|FLAG_WRITE:FLAG_WRITE);
    pData->hServer      = (iServer)? TAG_SERVER:TAG_TCP;


    #if SU_USE_IPV6

    memcpy(&pData->sAddr,pAddr,sizeof(pData->sAddr6));

    if(pData->sAddr.sin_family==AF_INET)
        {
        pData->sIpAddr.aAddr[0]=SU_IPV4_ID;
        pData->sIpAddr.aAddr[1]=pData->sAddr.sin_addr.s_addr;
        pData->usPort = htons  (pData->sAddr.sin_port);
        }
    else{
        memcpy(pData->sIpAddr.aAddr,pData->sAddr6.sin6_addr.s6_addr,16);
        pData->usPort = htons(pData->sAddr6.sin6_port);
        }

    #else

    memcpy(&pData->sAddr,pAddr,sizeof(pData->sAddr));

    pData->sIpAddr.aAddr[0] = pData->sAddr.sin_addr.s_addr;
    pData->usPort     = htons(pData->sAddr.sin_port);

    #endif


    #ifdef SU_WINDOWS

    pData->hEvent = WSACreateEvent();
    WSAEventSelect(pData->iSocket,pData->hEvent,(iServer)? FD_ACCEPT|FD_CLOSE:FD_READ|FD_WRITE|FD_CLOSE);

    #endif


    if(!iServer)
        {
        #if SEND_TIMEOUT>0
        TcpSendTimeout((TcpHandle)pData,SEND_TIMEOUT*1000);
        #endif

        #if RECV_TIMEOUT>0
        TcpRecvTimeout((TcpHandle)pData,RECV_TIMEOUT*1000);
        #endif
        }

    #if SU_HAS_SUBCALLS > 0
    pData->pSubCalls = 0;
    #endif



return (TcpHandle)pData;
}


//*****************************************************************************
//*
//*     TcpHandleFree
//*
//*****************************************************************************/
//  Releases a TCP handle
static inline void TcpHandleFree(TcpHandle hHandle)
{

    if(!hHandle)return;

    #ifdef SU_WINDOWS

    TcpData *pData=(TcpData*)hHandle;

    if(pData->hEvent)
        {
        WSACloseEvent(pData->hEvent);
        pData->hEvent = 0;
        }

    #endif

    FreeHandle(hHandle);
}

//*****************************************************************************
//*
//*     TcpHandleFake
//*
//*****************************************************************************/
//  Sets the local address and port
//  hHandle     : TCP handle of the socket
//  pAddr       : new ip address for the handle
//  iPort       : is the new port for the handle (0 means not change)
SUAPI void TcpHandleFake(TcpHandle hHandle,IpAddr *pAddr,int iPort)
{
TcpData *pData;


        pData = (TcpData*)hHandle;
    if(!pData)return;


    #if SU_USE_IPV6

    memcpy(&pData->sIpAddr.aAddr,pAddr->aAddr,sizeof(pData->sIpAddr.aAddr));
    if(iPort)pData->usPort  = iPort;

    #else

    pData->sIpAddr.aAddr[0] = pAddr->aAddr[0];
    if(iPort)pData->usPort  = iPort;

    #endif


}


//*****************************************************************************
//*
//*     TcpOpenCancel
//*
//*****************************************************************************
//  Cancel the open process of a tcp socket
//  pCancel     : handle to the open process
//  Returns 1 if cancel was done or 0 if not.
SUAPI int TcpOpenCancel(TcpCancel *pCancel)
{
int     iRet=0;



    SemaphoreLock(hGlobalSem);

    if(pCancel)
    if(pCancel->iValue==0x5A71D34E && pCancel->iSocket!=INVALID_SOCKET)
    if(pCancel->iSocket!=INVALID_SOCKET)
        {
        shutdown   (pCancel->iSocket,2);
        closesocket(pCancel->iSocket  );
        pCancel->iSocket = (int)INVALID_SOCKET;
        iRet=1;
        }

    SemaphoreUnlock(hGlobalSem);


return iRet;
}

//*****************************************************************************
//*
//*     TcpCancelDone
//*
//*****************************************************************************
//  Checks whether a cancel was done
//  pCancel     : handle to the open process
//  Returns 1 if cancel was done or 0 if not.
SUAPI int TcpCancelDone(TcpCancel *pCancel)
{
return (pCancel->iValue==0x5A71D34E)? 1:0;
}

//*****************************************************************************
//*
//*     TcpSetCancel
//*
//*****************************************************************************
//  Fill the TcpCancel structure (used for subcall initalization)
//  hHandle		: handle of the tcp socket
//  pCancel     : pointer to the cancel structure
//	iOn			: 1=init  0=exit
SUAPI int TcpSetCancel(TcpHandle hHandle,TcpCancel *pCancel,int iOn)
{
TcpData	   *pData=(TcpData*)hHandle;
int			iRet=1;




	SemaphoreLock(hGlobalSem);

	if(iOn)
		{
		pCancel->iSocket    = pData->iSocket;
		pCancel->iValue     = 0x5A71D34E;
		}
	else{
        if(pCancel->iSocket==INVALID_SOCKET)iRet=1;
		pCancel->iSocket    = (int)INVALID_SOCKET;
		pCancel->iValue     = 0x00000000;
		}

    SemaphoreUnlock(hGlobalSem);


return iRet;
}

//*****************************************************************************
//*
//*     TcpOpenWait
//*
//*****************************************************************************/
//  Creates a TCP socket
//  pAddr       : Is the IP adress.  i.e. IpAddress(&iAddr,"127.0.0.1",0);
//                If this pointer is zero an it is a sever socket the any
//                address will be used.
//  uPort       : TCP port number
//  iTimeout    : Is the timeout for the connect procces in ms. (0=no timeout)
//  pCancel     : Handle for the cancel operation (0=no cancel operation)
//  iServer     : Is the connection a sever (1) or a client (0)
//  iFlags      : Several Flags for the connection:
//                  TCP_BLOCKING    0x00000001  = sets the socket in the blocking mode
//                  TCP_NOBLOCKING  0x00000001  = sets the socket in the no blocking mode
//                  TCP_NODELAYACK  0x00000002  = disable nagle algorithm
//                  TCP_KEEP_ALIVE  0x00000004  = enable sending of keepalive packets
//                  TCP_REUSEADDR   0x00000008  = allows to reuse an used tcp port
//                  TCP_CLOSEDOWN   0x00000010  = close socket at TcpShutdown(...)
//                  TCP_DATABROKEN  0x00000020  = no broken signal if data received
//                  Bit 16..31                  = is the local port (0=auto, use TCP_USEPORT(a))
//  Retrns 0 at an error, or a TCP handle
SUAPI TcpHandle TcpOpenWait(IpAddr *pAddr,unsigned short uPort,int iTimeout,TcpCancel *pCancel,int iServer,int iFlags)
{
void                   *pData;
void                   *pTemp;
unsigned                uDiff;
unsigned                uStart;
fd_set                  sFdWait;
struct timeval          sTimeVal;
struct sockaddr_in      sAddr;
struct sockaddr_in      sTemp;
#if SU_USE_IPV6
struct sockaddr_in6_    sAddr6;
struct sockaddr_in6_    sTemp6;
int                     iMulti=0;
#endif
TcpHandle               hHandle;
SOCKET                  iSock;
IpAddr                  iTemp;
int                     iSize;
int                     iWait;
int                     iErr;
unsigned long           lNoneBlocking = 1;



    InitSocket();

    iSock     = INVALID_SOCKET;


    if(iServer)                                         // is it IPV4 any-sever-socket
        {
        iSock = CreateServerSocket(&pAddr,&iTemp,0);
        }
    else if(!pAddr)                                     // any address on clients not allowed
        {
        return 0;
        }


    #if SU_USE_IPV6
    if(pAddr->aAddr[0]!=SU_IPV4_ID)
        {                                               // open socket
        if(iSock==INVALID_SOCKET)iSock=socket(AF_INET6,SOCK_STREAM,IPPROTO_TCP);
        if(iSock==INVALID_SOCKET)return 0;

        memset(&sAddr6,0,sizeof(sAddr6));               // build local address
        memset(&sTemp6,0,sizeof(sTemp6));

        sAddr6.sin6_family      =  AF_INET6;
        sAddr6.sin6_port        =  htons(uPort);

        memcpy(sAddr6.sin6_addr.s6_addr,pAddr->aAddr,16);

        sTemp6.sin6_family      =  AF_INET6;
        sTemp6.sin6_port        =  (unsigned short)htons(iFlags>>16);

        pTemp = &sTemp6;
        pData = &sAddr6;
        iSize = sizeof(sAddr6);
        }
    else if(iSock!=INVALID_SOCKET)                      // IPV4 on dual socket
        {
        memset(&sAddr6,0,sizeof(sAddr6));               // build local address
        memset(&sTemp6,0,sizeof(sTemp6));

        sAddr6.sin6_family      =  AF_INET6;
        sAddr6.sin6_port        =  htons(uPort);

        memcpy(sAddr6.sin6_addr.s6_addr,IPV4TOV6,12);
        memcpy(sAddr6.sin6_addr.s6_addr+12,pAddr->aAddr+1,4);

        sTemp6.sin6_family      =  AF_INET6;
        sTemp6.sin6_port        =  (unsigned short)htons(iFlags>>16);

        pTemp = &sTemp6;
        pData = &sAddr6;
        iSize = sizeof(sAddr6);
        }
    else{
    #endif                                              // open socket
        if(iSock==INVALID_SOCKET)iSock=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
        if(iSock==INVALID_SOCKET)return 0;

        memset(&sAddr,0,sizeof(sAddr));                 // build local address
        memset(&sTemp,0,sizeof(sTemp));

        sAddr.sin_family      =  AF_INET;
        sAddr.sin_addr.s_addr =  GET_V4_ADDR(pAddr);
        sAddr.sin_port        =  htons(uPort);

        sTemp.sin_family      =  AF_INET;
        sTemp.sin_port        =  (unsigned short)htons(iFlags>>16);

        pTemp = &sTemp;
        pData = &sAddr;
        iSize = sizeof(sAddr);
    #if SU_USE_IPV6
        }
    #endif


    if(pCancel)                                         // set cancel handle
        {
        SemaphoreLock(hGlobalSem);
        pCancel->iSocket    = iSock;
        pCancel->iValue     = 0x5A71D34E;
        SemaphoreUnlock(hGlobalSem);
        }


    if(iFlags&TCP_REUSEADDR)                            // enable reuse of the address
        {
        int iVal;

        #if defined(RTTARGET) && !defined(LWIP)

        iVal=1;

        if(setsockopt(iSock,SOL_SOCKET,SO_REUSESOCK,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            closesocket(iSock);
            return 0;
            }

        #endif

        iVal=1;

        if(setsockopt(iSock,SOL_SOCKET,SO_REUSEADDR,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            closesocket(iSock);
            return 0;
            }
        }


    if(iServer)                                         // install a server socket
        {
            iErr = bind(iSock,(const struct sockaddr*)pData,iSize);
        if(!iErr)iErr = listen(iSock,8);
        #if SU_USE_IPV6
        if(!iErr && !iDualStack && iIpV6Allowed && IsIpAny(pAddr) && IsIpV6(pAddr))
            {
            iMulti = 1;
            }
        #endif
        }
    else{
        if(iFlags&0xFFFF0000)                           // use special local port
            {
            bind(iSock,(const struct sockaddr*)pTemp,iSize);
            }


        if(iTimeout || pCancel)
            {
            lNoneBlocking = 1;
                                                        // change socket into no blocking mode
            if(ioctlsocket(iSock,FIONBIO,&lNoneBlocking))
                {
                closesocket(iSock);
                return 0;
                }

            if(iTimeout    ==0)iTimeout=120000;
            if(iTimeout>120000)iTimeout=120000;
            if(iTimeout<     0)iTimeout=0;

            iWait = iTimeout;
            if(pCancel && iWait>200)iWait=200;

            sTimeVal.tv_sec  = iWait/1000;
            sTimeVal.tv_usec =(iWait%1000)*1000;

            uStart  = SysTickCount();
            iErr    = connect(iSock,(const struct sockaddr*)pData,iSize);

            for(;;)                                     // wait until connected
                {
                FD_ZERO(&sFdWait);
                FD_SET(iSock,&sFdWait);

                if(select(iSock+1,0,&sFdWait,0,&sTimeVal)>0)
                    {
                    if(iFlags&TCP_NOBLOCKING)
                        {
                        iErr = 0;
                        break;
                        }

                    lNoneBlocking = 0;                  // set back to in blocking mode
                    iErr = ioctlsocket(iSock,FIONBIO,&lNoneBlocking);
                    break;
                    }

                if(!pCancel || pCancel->iSocket!=(int)iSock)
                    {
                    iErr = SOCKET_ERROR;
                    break;
                    }

                   uDiff = SysTickCount()-uStart;
                if(uDiff>=(unsigned)iTimeout)           // check timeout
                    {
                    iErr = SOCKET_ERROR;
                    break;
                    }
                }
            }
        else{                                           // without timeout and cancel
            iErr = connect(iSock,(const struct sockaddr*)pData,iSize);
            }

        #ifdef LINUX

        if(iErr)SysSleep(50);                           // sometimes on linux connect(...) returns immeditly

        #endif
        }


    if(pCancel)                                         // reset cancel handle
        {
        SemaphoreLock(hGlobalSem);
        pCancel->iValue  = 0;
        SemaphoreUnlock(hGlobalSem);

        if(pCancel->iSocket==INVALID_SOCKET)            // socket cloesd by cancel function
            {
            return 0;
            }
        }


    if(iErr)                                            // can't open the socket
        {
        closesocket(iSock);
        return 0;
        }

    if(iFlags&TCP_NOBLOCKING)                           // set socket in none blocking mode
        {
        lNoneBlocking = 1;

        if(ioctlsocket(iSock,FIONBIO,&lNoneBlocking))
            {
            closesocket(iSock);
            return 0;
            }
        }

    if(iFlags&TCP_KEEP_ALIVE)                           // enable keepalive option
        {
        int iVal=1;

        if(setsockopt(iSock,SOL_SOCKET,SO_KEEPALIVE,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            closesocket(iSock);
            return 0;
            }
        }

    if(iFlags&TCP_NODELAYACK)                           // disable nagle algorithm
        {
        #ifndef RTTARGET

        int iVal = 1;

        if(setsockopt(iSock,SOL_SOCKET,TCP_NODELAY,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            closesocket(iSock);
            return 0;
            }

        #elif !defined(LWIP)

        int iVal = 0;

        if(setsockopt(iSock,SOL_SOCKET,SO_NAGLE,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            closesocket(iSock);
            return 0;
            }

        iVal = 0;

        if(setsockopt(iSock,SOL_SOCKET,SO_DELAYED_ACK,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            return 0;
            }

        #endif


        #ifdef LINUX

        iVal = 1;

        if(setsockopt(iSock,IPPROTO_TCP,TCP_QUICKACK,(const char*)&iVal,sizeof(iVal)==SOCKET_ERROR))
            {
            closesocket(iSock);
            return 0;
            }

        #endif
        }
	else if(iFlags&TCP_USENAGLE)						// enable the nagle algorithm
		{
        #ifndef RTTARGET

        int iVal = 0;

        if(setsockopt(iSock,SOL_SOCKET,TCP_NODELAY,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            return 0;
            }

        #elif !defined(LWIP)

        int iVal = 1;

        if(setsockopt(iSock,SOL_SOCKET,SO_NAGLE,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            return 0;
            }

        #endif
		}

    #ifndef RTTARGET

    else{                                               // disable nagle algorithm for fast answer

		int iVal = 1;

        setsockopt(iSock,SOL_SOCKET,TCP_NODELAY,(const char*)&iVal,sizeof(iVal));
        }

    #endif

    hHandle = TcpHandleAlloc(iSock,pData,iServer,iFlags);

    #if SU_USE_IPV6
    if(iMulti && hHandle)CreateTcpMulti(hHandle,uPort,iFlags);
    #endif


return hHandle;
}

//*****************************************************************************
//*
//*     TcpAccept
//*
//*****************************************************************************/
//  Accepts a connection on a server socket
//  hHandle : Handle of the sever socket
//  pAddr   : If this pointer is not zero, the ip address of the client will
//            stored there.
//  pPort   : If this pointer is not zero, the port number of the client will
//            stored there.
//  Returns a handle for the TCP connetion or 0 if no connetion is opened
SUAPI TcpHandle TcpAccept(TcpHandle hHandle,IpAddr *pAddr,unsigned short *pPort)
{
TcpData        *pData=(TcpData*)hHandle;
TcpHandle       hSock;
unsigned        uStart;
unsigned        uDiff;




    if(!hHandle)
        {
        return 0;
        }

    if(pData->ucSockFlags&TCP_NOBLOCKING)
        {
        return TcpAcceptWait(hHandle,pAddr,pPort,0);
        }

    for(;;)
        {
        uStart   = SysTickCount();

           hSock = TcpAcceptWait(hHandle,pAddr,pPort,12345678);
        if(hSock)break;

           uDiff = SysTickCount()-uStart;
        if(uDiff<500)SysSleep(200);
        }


    if(hSock==(TcpHandle)1)return NULL;



return hSock;
}

//*****************************************************************************
//*
//*     TcpWaitServer
//*
//*****************************************************************************/
//  Waits for an connection on a server socket with a timeout, the function
//  returns only if a connection is accepted, or the timeout is expired.
//  hHandle     : Handle of the sever socket
//  iTimeout    : Timeout in ms
//  Returns 2/1 if a connection is waiting
SUAPI int TcpWaitServer(TcpHandle hHandle,int iTimeout)
{
TcpData            *pData=(TcpData*)hHandle;
#if SU_USE_IPV6
TcpData            *pNext;
#endif
int                 iWait;
int                 iRet;



    if(!hHandle && !IS_TAG_SERVER(pData->hServer))      // check handle
        {
        SysSleep(iTimeout);
        return 0;
        }

    if(iTimeout==12345678)                              // hack for TcpAccept
        {
        iWait = 5000;
        }
    else{
        iWait = iTimeout;
        }


    #if SU_USE_IPV6

    if(IS_TAG_MULTI(pData->hServer))                    // is it a multiple port
        {
        pNext = (TcpData*)pData->hServer;
        }
    else{
        pNext = 0;
        }

    #endif

    #ifdef SU_WINDOWS

    if(iTimeout>0)
        {
        if(pData->ucConnFlags&FLAG_ACCEPT)              // try an unlocked accept
            {
            return 1;
            }

        #if SU_USE_IPV6

        if(pNext && (pNext->ucConnFlags&FLAG_ACCEPT))   // try the second multi server
            {
            return 1;
            }

        #endif
        }

    #endif

                                                        // wait until the next accept event
       iRet = SocketWaitAc(iWait,PARAM_WAIT(pData) PARAM_DUAL(pNext));
    if(iRet<=0)
        {
        return 0;
        }


    #if SU_USE_IPV6

    if(pNext && iRet==2)                                // is the multiple port signaled
        {
        return 2;
        }

    #endif



return 1;
}


//*****************************************************************************
//*
//*     TcpAcceptWait
//*
//*****************************************************************************/
//  Accepts a connection on a server socket with a timeout, the function
//  returns only if a connection is accepted, or the timeout is expired.
//  hHandle     : Handle of the sever socket
//  pAddr       : If this pointer is not zero, the ip address of the client will
//                stored there.
//  pPort       : If this pointer is not zero, the port number of the client will
//                stored there.
//  iTimeout    : Timeout in ms
//  Returns a handle for the TCP connetion or 0 if no connetion is opened
SUAPI TcpHandle TcpAcceptWait(TcpHandle hHandle,IpAddr *pAddr,unsigned short *pPort,int iTimeout)
{
TcpData                *pData=(TcpData*)hHandle;
TcpHandle               hError;
#if SU_USE_IPV6
TcpData                *pNext;
struct sockaddr_in6_    sAddr;
#else
struct sockaddr_in      sAddr;
#endif
int                     iWait;
int                     iSize;
int                     iSock;
int                     iRet;



    if(!hHandle)
        {
        SysSleep(iTimeout);
        return 0;
        }

    if(iTimeout==12345678)                              // hack for TcpAccept
        {
        iWait        = 5000;
        hError       = (TcpHandle)1;
        }
    else{
        iWait        = iTimeout;
        hError       = 0;
        }


    #if SU_USE_IPV6

    if(IS_TAG_MULTI(pData->hServer))                    // is it a multiple port
        {
        pNext = (TcpData*)pData->hServer;
        }
    else{
        pNext = 0;
        }

    #endif

    #ifdef SU_WINDOWS

    if(iTimeout>0)
        {
        if(pData->ucConnFlags&FLAG_ACCEPT)              // try an unlocked accept
            {
            iSize = sizeof(sAddr);
            iSock = accept(pData->iSocket,(struct sockaddr*)&sAddr,(LEN_TYPE)&iSize);

            if(iSock!=INVALID_SOCKET)
                {
                goto Done;
                }
            }

        #if SU_USE_IPV6

        if(pNext && (pNext->ucConnFlags&FLAG_ACCEPT))   // try the second multi server
            {
            iSize = sizeof(sAddr);
            iSock = accept(pNext->iSocket,(struct sockaddr*)&sAddr,(LEN_TYPE)&iSize);

            if(iSock!=INVALID_SOCKET)
                {
                pData = pNext;
                goto Done;
                }
            }

        #endif

        pData->ucConnFlags &= ~FLAG_ACCEPT;
        }

    #endif

                                                        // wait until the next accept event
       iRet = SocketWaitAc(iWait,PARAM_WAIT(pData) PARAM_DUAL(pNext));
    if(iRet<=0)
        {
        if(pAddr)SET_V4_ADDR(pAddr,0);
        return 0;
        }


    #if SU_USE_IPV6

    if(pNext && iRet==2)                                // is the multiple port signaled
        {
        pData = pNext;
        }

    #endif



    #if SU_USE_IPV6
    sAddr.sin6_family = IsIpV6(&pData->sIpAddr)? AF_INET6:AF_INET;
    #endif

       iSize = sizeof(sAddr);
       iSock = accept(pData->iSocket,(struct sockaddr*)&sAddr,(LEN_TYPE)&iSize);
    if(iSock==INVALID_SOCKET)
        {
        if(pAddr)SET_V4_ADDR(pAddr,0);
        return hError;
        }

    #ifdef  SU_WINDOWS
Done:
    #endif


    #ifndef RTTARGET

	int iVal = 1;										// disable nagle algorithm for fast answer

    setsockopt(iSock,SOL_SOCKET,TCP_NODELAY,(const char*)&iVal,sizeof(iVal));

    #endif


    #if SU_USE_IPV6

    if(sAddr.sin6_family==AF_INET)
        {
        if(pAddr)SET_V4_ADDR(pAddr,((struct sockaddr_in*)&sAddr)->sin_addr.s_addr);
        if(pPort)*pPort = htons   (((struct sockaddr_in*)&sAddr)->sin_port);
        }
    else{
        if(pAddr)memcpy(pAddr->aAddr,sAddr.sin6_addr.s6_addr,16);
        if(pPort)*pPort=htons(sAddr.sin6_port);
        }

    #else

    if(pAddr)SET_V4_ADDR(pAddr,sAddr.sin_addr.s_addr);
    if(pPort)*pPort=htons(sAddr.sin_port);

    #endif



return TcpHandleAlloc(iSock,&sAddr,TCP_CLIENT,0);
}

//*****************************************************************************
//*
//*     TcpSetFlags
//*
//*****************************************************************************
//  Changes the flags of a TCP socket
//  hHandle     : TCP handle of the socket
//  iFlags      : Several Flags for the connection:
//                  TCP_NOBLOCKING  0x00000001  = sets the socket in the no blocking mode
//                  TCP_NODELAYACK  0x00000002  = disable nagle algorithm
//                  TCP_KEEP_ALIVE  0x00000004  = enable sending of keepalive packets
//                  TCP_CLOSEDOWN   0x00000010  = close socket at TcpShutdown(...)
//                  TCP_DATABROKEN  0x00000020  = no broken signal if data received
//					TCP_USENAGLE	0x00000040	= enable the nagle algorithm
//  Returns 0 if an error occurs or 1 if all ok
SUAPI int TcpSetFlags(TcpHandle hHandle,int iFlags)
{
TcpData        *pData=(TcpData*)hHandle;


    if(!pData)return 0;


    if(iFlags&TCP_NOBLOCKING)                           // set socket in none blocking mode
        {
        unsigned long   lNoneBlocking = 1;

        if(ioctlsocket(pData->iSocket,FIONBIO,&lNoneBlocking))
            {
            return 0;
            }
        }

    if(iFlags&TCP_KEEP_ALIVE)                           // enable keepalive option
        {
        int iVal=1;

        if(setsockopt(pData->iSocket,SOL_SOCKET,SO_KEEPALIVE,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            return 0;
            }
        }

    if(iFlags&TCP_NODELAYACK)                           // disable nagle algorithm
        {
        #ifndef RTTARGET

        int iVal = 1;

        if(setsockopt(pData->iSocket,SOL_SOCKET,TCP_NODELAY,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            return 0;
            }

        #elif !defined(LWIP)

        int iVal = 0;

        if(setsockopt(pData->iSocket,SOL_SOCKET,SO_NAGLE,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            return 0;
            }

        iVal = 0;

        if(setsockopt(pData->iSocket,SOL_SOCKET,SO_DELAYED_ACK,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            return 0;
            }

        #endif

        #ifdef LINUX

        iVal = 1;

        if(setsockopt(pData->iSocket,IPPROTO_TCP,TCP_QUICKACK,(const char *)&iVal,sizeof(iVal)==SOCKET_ERROR))
            {
            return 0;
            }

        #endif
        }

	if(iFlags&TCP_USENAGLE)								// enable the nagle algorithm
		{
        #ifndef RTTARGET

        int iVal = 0;

        if(setsockopt(pData->iSocket,SOL_SOCKET,TCP_NODELAY,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            return 0;
            }

        #elif !defined(LWIP)

        int iVal = 1;

        if(setsockopt(pData->iSocket,SOL_SOCKET,SO_NAGLE,(const char*)&iVal,sizeof(iVal))==SOCKET_ERROR)
            {
            return 0;
            }

        #endif
		}

                                                        // change the flags
    pData->ucSockFlags |= (unsigned char)(iFlags&(TCP_NOBLOCKING|TCP_NODELAYACK|TCP_KEEP_ALIVE|TCP_DATABROKEN|TCP_CLOSEDOWN));



return 1;
}

//*****************************************************************************
//*
//*     TcpBlock
//*
//*****************************************************************************/
//  Changes the blocking mode of a TCP socket
//  hHandle     : TCP handle of the socket
//  iOn         : 0=none-blocking  1=blocking
//  Returns 0 if an error occurs or 1 if all ok
SUAPI int TcpBlock(TcpHandle hHandle,int iOn)
{
TcpData        *pData=(TcpData*)hHandle;


    if(!pData)return 0;

    if(!iOn && !(pData->ucSockFlags&TCP_NOBLOCKING))
        {
        #ifndef SU_WINDOWS                              // windows has always noneblocking sockets
        unsigned long  lNoneBlocking = 1;
        if(ioctlsocket(pData->iSocket,FIONBIO,&lNoneBlocking))return 0;
        #endif

        pData->ucSockFlags |= TCP_NOBLOCKING;
        }

    if(iOn &&  (pData->ucSockFlags&TCP_NOBLOCKING))
        {
        #ifndef SU_WINDOWS                              // windows has always noneblocking sockets
        unsigned long  lNoneBlocking = 0;
        if(ioctlsocket(pData->iSocket,FIONBIO,&lNoneBlocking))return 0;
        #endif

        pData->ucSockFlags &= ~TCP_NOBLOCKING;
        }


return 1;
}

//*****************************************************************************
//*
//*     TcpCloseDown
//*
//*****************************************************************************/
//  Changes the close down mode of a TCP socket
//  hHandle     : TCP handle of the socket
//  iOn         : 0=not close at shutdown 1=close at shutdown
//  Returns 0 if an error occurs or 1 if all ok
SUAPI int TcpCloseDown(TcpHandle hHandle,int iOn)
{
TcpData        *pData=(TcpData*)hHandle;


    if(!pData)return 0;

    if(iOn)pData->ucConnFlags|= FLAG_DOWN;
    else   pData->ucConnFlags&=~FLAG_DOWN;


return 1;
}

//*****************************************************************************
//*
//*     TcpSetBuffers
//*
//*****************************************************************************/
//  Changes the send and receive buffer sizes of a socket
//  hHandle     : TCP handle of the socket
//  iRecvSize   : Size for the receive buffer (0=none-change)
//  iSendSize   : Size for the send    buffer (0=none-change)
//  Returns     Bit0 :  0=receive buffer not change  1=changed
//              Bit1 :  0=send    buffer not change  1=changed
SUAPI int TcpSetBuffers(TcpHandle hHandle,int iRecvSize,int iSendSize)
{
TcpData        *pData=(TcpData*)hHandle;
int             iRet=0;


    if(!pData)return 0;


    #ifndef RTTARGET

    if(iRecvSize && setsockopt(pData->iSocket,SOL_SOCKET,SO_RCVBUF,(char*)&iRecvSize,sizeof(iRecvSize)) != -1)
        {
        iRet |=1;
        }

    if(iSendSize && setsockopt(pData->iSocket,SOL_SOCKET,SO_SNDBUF,(char*)&iSendSize,sizeof(iSendSize)) != -1)
        {
        iRet |=2;
        }

    #endif



return iRet;
}

//*****************************************************************************
//*
//*     TcpSendTimeout
//*
//*****************************************************************************
//  Sets the send timeout of a socket.
//  hHandle     : TCP handle from the socket
//  iTimeout    : Is the timeout in milliseconds
//  Return the 0 if ok, or -1 if an error occurs
int TcpSendTimeout(TcpHandle hHandle, int iTimeout)
{
TcpData    *pData  = (TcpData*)hHandle;
int         iError;


    if(!pData)return -1;


    if(iTimeout<500)
        {
        iTimeout = 500;
        }


    #if defined(RTTARGET) && !defined(LWIP)
        {
        struct timeval  sTime;

        sTime.tv_sec    =  iTimeout/1000;
        sTime.tv_usec   = (iTimeout%1000)*1000;

        iError = setsockopt(pData->iSocket,SOL_SOCKET,SO_SEND_TIMEO,(const char*)&sTime,sizeof(sTime));
        }
    #elif defined(SU_WINDOWS) || defined(LWIP)
        {
        int     iVal;

        iVal = iTimeout;

        iError = setsockopt(pData->iSocket,SOL_SOCKET,SO_SNDTIMEO,(const char*)&iVal,sizeof(iVal));
        }
    #elif defined(LINUX)
        {
        struct timeval  sTime;

        sTime.tv_sec    =  iTimeout/1000;
        sTime.tv_usec   = (iTimeout%1000)*1000;

        iError = setsockopt(pData->iSocket,SOL_SOCKET,SO_SNDTIMEO,(const char*)&sTime,sizeof(sTime));
        }
    #else

    #error  wrong operating system

    #endif



return iError;
}


//*****************************************************************************
//*
//*     TcpRecvTimeout
//*
//*****************************************************************************
//  Sets the receive timeout of a socket.
//  hHandle     : TCP handle from the socket
//  iTimeout    : Is the timeout in milliseconds
//  Return the 0 if ok, or -1 if an error occurs
int TcpRecvTimeout(TcpHandle hHandle, int iTimeout)
{
TcpData    *pData  = (TcpData*)hHandle;
int         iError;


    if(!pData)return -1;

    if(iTimeout<500)
        {
        iTimeout = 500;
        }



    #if defined(RTTARGET) && !defined(LWIP)
        {
        struct timeval  sTime;

        sTime.tv_sec    =  iTimeout/1000;
        sTime.tv_usec   = (iTimeout%1000)*1000;

        iError = setsockopt(pData->iSocket,SOL_SOCKET,SO_RCV_TIMEO,(const char*)&sTime,sizeof(sTime));
        }
    #elif defined(SU_WINDOWS) || defined(LWIP)
        {
        int     iVal;

        iVal = iTimeout;

        iError = setsockopt(pData->iSocket,SOL_SOCKET,SO_RCVTIMEO,(const char*)&iVal,sizeof(iVal));
        }
    #elif defined(LINUX)
        {
        struct timeval  sTime;

        sTime.tv_sec    =  iTimeout/1000;
        sTime.tv_usec   = (iTimeout%1000)*1000;

        iError = setsockopt(pData->iSocket,SOL_SOCKET,SO_RCVTIMEO,(const char*)&sTime,sizeof(sTime));
        }
    #endif


return iError;
}


//*****************************************************************************
//*
//*     TcpShutdown
//*
//*****************************************************************************/
//  Disconnects a TCP socket. After this call no data will send or receive
//  on the socket
//  hHandle     : Handle of the TCP socket
//  Returns 0 if an error occurs or 1 if all ok
SUAPI int TcpShutdown(TcpHandle hHandle)
{
TcpData    *pData=(TcpData*)hHandle;
int         iSocket;

    if(!hHandle)return 0;
    if(shutdown(pData->iSocket,2))return 0;

    if(pData->ucConnFlags&FLAG_DOWN)
        {
        iSocket            = pData->iSocket;
        pData->ucConnFlags = (char)FLAG_BROKEN;
        pData->iSocket     = -1;
        closesocket(iSocket);
        }


return 1;
}

//*****************************************************************************
//*
//*     TcpClose
//*
//*****************************************************************************/
//  Closes a TCP socket TCP socket
//  hHandle     : Handle of the TCP socket
//  Returns 0 if an error occurs or 1 if all ok
SUAPI int TcpClose(TcpHandle hHandle)
{
TcpData  *pData=(TcpData*)hHandle;


    if(!hHandle || !IS_TAG_TCP(pData->hServer))
        {
        return 0;
        }

    #if SU_HAS_SUBCALLS

    TcpSubCalls    *pSubCalls;
    unsigned        uLevel;

    if(pData->pSubCalls)                                // exists subcalls
        {
        pSubCalls  = pData->pSubCalls;
        uLevel     = pData->iSubLevels;
        pSubCalls += uLevel;

        pData->pSubCalls = 0;

        for(;uLevel>0;uLevel--,pSubCalls--)
            {
            if(!pSubCalls->pClose)continue;
            pSubCalls->pClose(pSubCalls->pHandle);
            }

        FreeHandle(pSubCalls);
        }

    #endif

    #if SU_USE_IPV6
    if(IS_TAG_MULTI(pData->hServer))                    // exist a second server socket
        {
        TcpClose((TcpHandle)pData->hServer);
        }
    #endif

    #ifndef LWIP
    shutdown(pData->iSocket,2);
    #endif

    closesocket(pData->iSocket);
    pData->hServer = 0;
    TcpHandleFree(hHandle);


return 1;
}


//*****************************************************************************
//*
//*     TcpPut
//*
//*****************************************************************************/
//  Sends data on a TCP socket
//  hHandle     : Handle of the TCP socket
//  pMsg        : Pointer to the data buffer
//  iLen        : Size of the data buffer
//  Returns the count of the sended bytes, or zero if an error occurs.
SUAPI int TcpPut(TcpHandle hHandle,const void *pMsg,int iLen)
{
#ifdef  SU_WINDOWS
unsigned    uStop=GetTickCount();
int         iWait=0;
#endif
TcpData    *pData=(TcpData*)hHandle;
int         iDiff;
int         iRet;
int         iNum;
int         iPos;
int         iErr;


    if(!hHandle)                                        // check handle
        {
        return 0;
        }

    #if SU_HAS_SUBCALLS

    if(pData->pSubCalls)                                // exists subcalls
        {
        if(TcpIsBrokenEx(hHandle,1))
            {
            return 0;
            }

        iRet = TcpSubCallSend(pData->pSubCalls,pData->iSubLevels,iLen,(char*)pMsg);

        return (iRet<0)? 0:iRet;
        }

    #endif

    for(iPos=0,iErr=0;iPos<iLen;)
        {
        iNum = iLen-iPos;
                                                        // send data
           iRet = send(pData->iSocket,(char*)pMsg+iPos,iNum,SEND_FLAGS);
        if(iRet<0)
            {
            if(TcpIsBrokenEx(hHandle,1))break;
            iRet = 0;
            }
        else{
            iPos += iRet;
            if(iPos>=iLen)break;                        // all bytes send
            if(TcpIsBrokenEx(hHandle,1))break;
            }

        if(pData->ucSockFlags&TCP_NOBLOCKING)           // not block the call
            {
            break;
            }

        pData->ucConnFlags &= ~FLAG_WRITE;

        if(iRet>0)
            {
            iErr = 0;
            }
        else{
            #ifdef  SU_WINDOWS

            if(!iWait)                                  // get send timeout of socket
                {
                iDiff = sizeof(int);
                getsockopt(pData->iSocket,SOL_SOCKET,SO_SNDTIMEO,(char*)&iWait,&iDiff);
                if(!iWait)iWait=(SEND_TIMEOUT)? SEND_TIMEOUT:-1;
                uStop  += iWait;
                }

            iErr++;

            if(iWait>0)                                 // absolute timeout
                {
                   iDiff = uStop-GetTickCount();
                if(iDiff<=100)break;

                iDiff = 250;
                }
            else{                                       // no timeout defined
                if(iErr>10*4)break;
                }

            #else

               iErr++;
            if(iErr>2)break;

            iDiff = 250;

            #endif
            }

        iRet = SocketWaitWr(iDiff,PARAM_WAIT(pData));   // wait until write is possible

        if(iRet<0)
            {
            break;
            }
        }


return iPos;
}

//*****************************************************************************
//*
//*     TcpPutSub
//*
//*****************************************************************************/
//  Sends data on a TCP socket for subcalls
//  hHandle     : Handle of the TCP socket
//  pMsg        : Pointer to the data buffer
//  iLen        : Size of the data buffer
//  Returns the count of the sended bytes, or zero if an error occurs.
#if SU_HAS_SUBCALLS
static int TcpPutSub(TcpHandle hHandle,const void *pMsg,int iLen)
{
#ifdef  SU_WINDOWS
unsigned    uStop=GetTickCount();
int         iWait=0;
#endif
TcpData    *pData=(TcpData*)hHandle;
int         iDiff;
int         iRet;
int         iNum;
int         iPos;
int         iErr;


    if(!hHandle)                                        // check handle
        {
        return 0;
        }


    for(iPos=0,iErr=0;iPos<iLen;)
        {
        iNum = iLen-iPos;
                                                        // send data
           iRet = send(pData->iSocket,(char*)pMsg+iPos,iNum,SEND_FLAGS);
        if(iRet<0)
            {
            if(TcpIsBrokenEx(hHandle,1))break;
            iRet = 0;
            }
        else{
            iPos += iRet;
            if(iPos>=iLen)break;                        // all bytes send
            if(TcpIsBrokenEx(hHandle,1))break;
            }

        pData->ucConnFlags &= ~FLAG_WRITE;

        if(iRet>0)
            {
            iErr = 0;
            }
        else{
            #ifdef  SU_WINDOWS

            if(!iWait)                                  // get send timeout of socket
                {
                iDiff = sizeof(int);
                getsockopt(pData->iSocket,SOL_SOCKET,SO_SNDTIMEO,(char*)&iWait,&iDiff);
                if(!iWait)iWait=(SEND_TIMEOUT)? SEND_TIMEOUT:-1;
                uStop  += iWait;
                }

            iErr++;

            if(iWait>0)                                 // absolute timeout
                {
                   iDiff = uStop-GetTickCount();
                if(iDiff<=100)break;

                iDiff = 250;
                }
            else{                                       // no timeout defined
                if(iErr>10*4)break;
                }

            #else

               iErr++;
            if(iErr>2)break;

            iDiff = 250;

            #endif
            }

        iRet = SocketWaitWr(iDiff,PARAM_WAIT(pData));   // wait until write is possible

        if(iRet<0)
            {
            break;
            }
        }


return iPos;
}
#endif

//*****************************************************************************
//*
//*     TcpGet
//*
//*****************************************************************************/
//  Reveives data from a TCP socket. If the socket is none blocking the
//  function returns immediately. On blocking sockets, the function waits
//  if all bytes received.
//  hHandle     : Handle of the TCP socket
//  pMsg        : Pointer to the data buffer
//  iLen        : Size of the data buffer
//  Returns -1 if the connection is brocken or no byte is received on blocking sockets
//  Returns  0 on NONE_BLOCKING sockets if no byte is received.
//  Returns the count of the received bytes
SUAPI int TcpGet(TcpHandle hHandle,void *pMsg,int iLen)
{
TcpData    *pData=(TcpData*)hHandle;
int         iBytes;


    if(!hHandle)return -1;


    #if SU_HAS_SUBCALLS

    if(pData->pSubCalls)                                // exists a subcall array
        {
        unsigned char  *pPtr=(unsigned char*)pMsg;
        TcpSubCalls    *pSubCalls;
        unsigned        uStop;
        int             iWait;
        int             iRet;

        pSubCalls   = pData->pSubCalls;
        pSubCalls  += pData->iSubLevels;                // goto last subcall
        iBytes      = pSubCalls->pCountOut(pSubCalls->pHandle);
        iRet        = 0;

        if(iLen<0)iLen=0;

        uStop       = SysTickCount();
        uStop      += (RECV_TIMEOUT)? RECV_TIMEOUT*1000:30000;

        for(;;)
            {                                           // count of bytes in output buffer
            if(iBytes>iLen)iBytes=iLen;
               iBytes = pSubCalls->pOutRd(pSubCalls->pHandle,pPtr,iLen);
            if(iBytes>0)
                {
                pPtr += iBytes;
                iRet += iBytes;
                iLen -= iBytes;
                }

            if(iLen<=0 || (pData->ucSockFlags&TCP_NOBLOCKING))
                {
                break;
                }

               iWait = uStop-SysTickCount();
            if(iWait<=0)break;

               iBytes = TcpWait(hHandle,iWait);
            if(iBytes<0)break;
            }

        if(iLen==0 && iRet==0)                          // if no data then check for broken connection
            {
            if(!(pData->ucSockFlags&TCP_NOBLOCKING) || TcpIsBroken(hHandle))
                {
                return -1;
                }
            }

        return iRet;
        }

    #endif



    if(pData->ucSockFlags&TCP_NOBLOCKING)
        {
        if(SocketWaitRd(0,PARAM_WAIT(pData))<=0)        // data in receive buffer
            {
            return 0;
            }

        if(iLen<=0)                                     // empty buffer
            {
            if(TcpIsBroken(hHandle))return -1;
            return 0;
            }

        iBytes = recv(pData->iSocket,(char*)pMsg,iLen,RECV_FLAGS);

        if(iBytes<0)return 0;
        }
    else{
        if(iLen<=0)                                     // empty buffer
            {
            if(TcpIsBroken(hHandle))return -1;
            return 0;
            }

        #ifdef SU_WINDOWS
                                                        // windows has always none blocking sockets
        int         iSize=sizeof(int);
        int         iWait=0;

        getsockopt(pData->iSocket,SOL_SOCKET,SO_RCVTIMEO,(char*)&iWait,&iSize);
        if(!iWait)iWait=(RECV_TIMEOUT)? RECV_TIMEOUT*1000:30000;

        SocketWaitRd(iWait,PARAM_WAIT(pData),iLen);

        #endif


        iBytes = recv(pData->iSocket,(char*)pMsg,iLen,RECV_FLAGS);

        if(iBytes<0)return -1;
        }


return iBytes;
}

//*****************************************************************************
//*
//*     TcpGetSub
//*
//*****************************************************************************/
//  Reveives data from a TCP socket. If the socket is none blocking the
//  function returns immediately. On blocking sockets, the function waits
//  if all bytes received. (for subcall)
//  hHandle     : Handle of the TCP socket
//  pMsg        : Pointer to the data buffer
//  iLen        : Size of the data buffer
//  Returns -1 if the connection is brocken or no byte is received on blocking sockets
//  Returns  0 on NONE_BLOCKING sockets if no byte is received.
//  Returns the count of the received bytes
#if SU_HAS_SUBCALLS
static int TcpGetSub(TcpHandle hHandle,void *pMsg,int iLen)
{
TcpData    *pData=(TcpData*)hHandle;
int         iBytes;


    if(!hHandle)return -1;

    iBytes = recv(pData->iSocket,(char*)pMsg,iLen,RECV_FLAGS);


return iBytes;
}
#endif

//*****************************************************************************
//*
//*     TcpGetWait
//*
//*****************************************************************************/
//  Receives data from a TCP socket with a timeout, the function returns only
//  if all data bytes are receives, or the timeout is expired.
//  hHandle     : Handle of the TCP socket
//  pMsg        : Pointer to the data buffer
//  iLen        : Count of the bytes you will receive
//  iTimeout    : iTimeout in millisecounds
//  Returns -1 if the connection is brocken or no byte is received on blocking sockets
//  Returns  0 on NONE_BLOCKING sockets if no byte is received.
//  Returns the count of the received bytes
SUAPI int TcpGetWait(TcpHandle hHandle,void *pMsg,int iLen,int iTimeout)
{
TcpData        *pData=(TcpData*)hHandle;
SysDW           uStart;
int             iBytes;
int             iWait;
int             iNum;


    if(hHandle==0)                                      // check handle
        {
        return -1;
        }


    #if SU_HAS_SUBCALLS

    if(pData->pSubCalls)                                // exists a subcall array
        {
        unsigned char  *pPtr=(unsigned char*)pMsg;
        TcpSubCalls    *pSubCalls;
        unsigned        uStop;
        int             iRet;

        pSubCalls   = pData->pSubCalls;
        pSubCalls  += pData->iSubLevels;                // goto last subcall
        iBytes      = pSubCalls->pCountOut(pSubCalls->pHandle);
        iRet        = 0;

        if(iLen<0)iLen=0;

        uStop       = SysTickCount()+iTimeout;

        for(;;)
            {                                           // count of bytes in output buffer
            if(iBytes>iLen)iBytes=iLen;
               iBytes = pSubCalls->pOutRd(pSubCalls->pHandle,pPtr,iLen);
            if(iBytes>0)
                {
                pPtr += iBytes;
                iRet += iBytes;
                iLen -= iBytes;
                }

            if(iLen<=0)
                {
                break;
                }

               iTimeout = uStop-SysTickCount();
            if(iTimeout<=0)break;

               iBytes = TcpWait(hHandle,iTimeout);
            if(iBytes<0)break;
            }

        if(iLen==0 && iRet==0)                          // if no data then check for broken connection
            {
            if(TcpIsBroken(hHandle))
                {
                return -1;
                }
            }

        return iRet;
        }

    #endif



    if(iTimeout<0)                                      // check timeout
        {
        if(TcpIsBroken(hHandle))return -1;
        return 0;
        }

    iWait  = iTimeout;
    uStart = SysTickCount();

    for(iBytes=0;iBytes<iLen;)
        {
        if(SocketWaitRd(iWait,PARAM_WAIT(pData))<=0)    // wait until data received
            {
            if(TcpIsBroken(hHandle))                    // check for broken connection
                {
                return (!iBytes)? -1:iBytes;
                }

               iWait=iTimeout-(SysTickCount()-uStart);
            if(iWait<=0)return iBytes;

            SysSleep((iWait>300)? 30:10);
            continue;
            }

        if(iLen==iBytes)                                // empty buffer
            {
            iNum = 0;
            }
        else{
               iNum = recv(pData->iSocket,(char*)pMsg+iBytes,iLen-iBytes,RECV_FLAGS);
            if(iNum<0)return -1;
            }

        if(iNum==0)                                     // check for broken connection
            {
            if(!(pData->ucSockFlags&TCP_NOBLOCKING) || TcpIsBroken(hHandle))
                {
                return (!iBytes)? -1:iBytes;
                }
            }


        iBytes += iNum;

           iWait =iTimeout-(SysTickCount()-uStart);
        if(iWait<=0)return iBytes;
        if(iNum ==0)SysSleep((iWait>300)? 30:10);       // BUG in Windows ip stack !!!
        }


return iBytes;
}


//*****************************************************************************
//*
//*     TcpGetNext
//*
//*****************************************************************************/
//  Receives the next data from a TCP socket with a timeout, the function
//  returns if one or more data bytes are receives, or the timeout time is expired.
//  hHandle     : Handle of the TCP socket
//  pMsg        : Pointer to the data buffer
//  iLen        : Count of the bytes you will maximal receive
//  iTimeout    : Timeout in millisecounds
//  Returns -1 if the connection is brocken or no byte is received on blocking sockets
//  Returns  0 on timeouts.
//  Returns the count of the received bytes
SUAPI int TcpGetNext(TcpHandle hHandle,void *pMsg,int iLen,int iTimeout)
{
TcpData        *pData=(TcpData*)hHandle;
unsigned long   uReady;
SysDW           uStart;
int             iBytes;
int             iWait;



    if(hHandle==0)                                      // check handle
        {
        return -1;
        }


    #if SU_HAS_SUBCALLS

    if(pData->pSubCalls)                                // exists a subcall array
        {
        unsigned char  *pPtr=(unsigned char*)pMsg;
        TcpSubCalls    *pSubCalls;
        unsigned        uStop;
        int             iRet;

        pSubCalls   = pData->pSubCalls;
        pSubCalls  += pData->iSubLevels;                // goto last subcall
        iBytes      = pSubCalls->pCountOut(pSubCalls->pHandle);
        iRet        = 0;

        if(iLen<0)iLen=0;

        uStop       = SysTickCount()+iTimeout;

        for(;;)
            {                                           // count of bytes in output buffer
            if(iBytes>iLen)iBytes=iLen;
               iBytes = pSubCalls->pOutRd(pSubCalls->pHandle,pPtr,iLen);
            if(iBytes>0)
                {
                pPtr += iBytes;
                iRet += iBytes;
                iLen -= iBytes;
                break;
                }

            if(iLen<=0)
                {
                break;
                }

               iTimeout = uStop-SysTickCount();
            if(iTimeout<=0)break;

               iBytes = TcpWait(hHandle,iTimeout);
            if(iBytes<0)break;
            }

        if(iLen==0 && iRet==0 && TcpIsBroken(hHandle))  // if no data then check for broken connection
            {
            return -1;
            }

        return iRet;
        }

    #endif



    if(iTimeout<0)                                      // check timeout
        {
        if(TcpIsBroken(hHandle))return -1;
        return 0;
        }

    iWait  = iTimeout;
    uStart = SysTickCount();

    for(;;)
        {
        if(SocketWaitRd(iWait,PARAM_WAIT(pData))<=0)    // wait until data received
            {
               iWait=iTimeout-(SysTickCount()-uStart);
            if(iWait<=0)return 0;

            SysSleep((iWait>300)? 30:10);

            continue;
            }

        if(ioctlsocket(pData->iSocket,FIONREAD,&uReady))// get byte count in queue
            {
            return -1;
            }

        if((int)uReady>iLen)uReady=iLen;

        if(uReady==0)                                   // empty buffer
            {
            if(TcpIsBroken(hHandle))                    // check for broken connection
                {
                return -1;
                }

            iBytes = 0;
            }
        else{
            iBytes = recv(pData->iSocket,(char*)pMsg,uReady,RECV_FLAGS);
            }

        if(iBytes>0)break;
        if(iBytes<0)
            {
            return -1;
            }

        if(!(pData->ucSockFlags&TCP_NOBLOCKING))        // is it a blocking socket
            {
            return -1;
            }


        if(TcpIsBroken(hHandle))                        // check for broken connection
            {
            return -1;
            }


           iWait =iTimeout-(SysTickCount()-uStart);
        if(iWait<=0)return 0;

        SysSleep((iWait>300)? 30:10);                   // BUG in Windows ip stack !!!
        }


return iBytes;
}

//*****************************************************************************
//*
//*     TcpWait
//*
//*****************************************************************************/
//  Waits until the next data from a TCP socket was received.
//  hHandle     : Handle of the TCP socket
//  iTimeout    : Timeout in millisecounds
//  Returns -1 if the connection is brocken
//  Returns  0 on timeouts.
//  Returns  the count of data in the the input buffer
SUAPI int TcpWait(TcpHandle hHandle,int iTimeout)
{
TcpData        *pData=(TcpData*)hHandle;
unsigned long   uReady;



    if(hHandle==0)                                      // check handle
        {
        return -1;
        }

    if(iTimeout<0)                                      // check timeout
        {
        if(TcpIsBroken(hHandle))return -1;
        return 0;
        }

    #if SU_HAS_SUBCALLS

	unsigned    uFlag=0;
    unsigned    uStop;

	if((iTimeout&0xFFFF0000)==SU_SUB_TIMEOUT)			// timeouts only for sub calls
		{
		iTimeout &= 0xFFFF;
		uFlag     = 1;
		}

    if(pData->pSubCalls)                                // exists a subcall array
        {
        TcpSubCalls    *pSubCalls;

        if(!ioctlsocket(pData->iSocket,FIONREAD,&uReady) && uReady)
            {
            TcpSubCallRecv(pData->pSubCalls,pData->iSubLevels,uReady);
            }

        uStop       = SysTickCount()+iTimeout;
        pSubCalls   = pData->pSubCalls;
        pSubCalls  += pData->iSubLevels;                // goto last subcall
        uReady      = pSubCalls->pCountOut(pSubCalls->pHandle);

        if(uReady)return uReady;
        }

Retry:

    #endif


    if(SocketWaitRd(iTimeout,PARAM_WAIT(pData))<=0)     // wait until data received
        {
        if(TcpIsBroken(hHandle))return -1;              // check for broken connection
        return 0;
        }

    if(ioctlsocket(pData->iSocket,FIONREAD,&uReady))    // get byte count in queue
        {
        return -1;
        }


    #if SU_HAS_SUBCALLS

    if(pData->pSubCalls && uReady)                      // exists a subcall array
        {
        TcpSubCalls    *pSubCalls;

        TcpSubCallRecv(pData->pSubCalls,pData->iSubLevels,uReady);

		if(uFlag && uReady)								// stop on subcall receive
			{
			return 0;
			}

        pSubCalls   = pData->pSubCalls;
        pSubCalls  += pData->iSubLevels;                // goto last subcall
        uReady      = pSubCalls->pCountOut(pSubCalls->pHandle);

        if(uReady==0)                                   // no output data after
            {
               iTimeout = uStop-SysTickCount();
            if(iTimeout>0)
                {
                goto Retry;
                }
            }
        }

    #endif


return uReady;
}


//*****************************************************************************
//*
//*     TcpWaitSend
//*
//*****************************************************************************/
//  Waits until data could be send on a TCP socket.
//  hHandle     : Handle of the TCP socket
//  iTimeout    : Timeout in millisecounds
//  Returns -1 if the connection is brocken, 0 on timeouts,
//  or 1 if it is possible to sending data.
SUAPI int TcpWaitSend(TcpHandle hHandle,int iTimeout)
{
TcpData        *pData=(TcpData*)hHandle;



    if(hHandle==0)                                      // check handle
        {
        return -1;
        }

    if(iTimeout<0)                                      // check timeout
        {
        if(TcpIsBrokenEx(hHandle,1))return -1;
        return 0;
        }

    if(SocketWaitWr(iTimeout,PARAM_WAIT(pData))<=0)     // wait until data received
        {
        if(TcpIsBrokenEx(hHandle,1))return -1;          // check for broken connection
        return 0;
        }



return 1;
}


//*****************************************************************************
//*
//*     TcpGetCount
//*
//*****************************************************************************/
//  Gets the size of stored data in the receive buffer of the socket.
//  hHandle : TCP handle from the socket
//  Return the size or 0 if no data is in the buffer, or -1 on an error occurs.
SUAPI int TcpGetCount(TcpHandle hHandle)
{
TcpData        *pData=(TcpData*)hHandle;
unsigned long   uDataReady;


    if(!hHandle)return 0;
    if(ioctlsocket(pData->iSocket,FIONREAD,&uDataReady))        // get bytes to read
        {
        return -1;
        }


    #if SU_HAS_SUBCALLS

    if(pData->pSubCalls)                                        // exists a subcall array
        {
        TcpSubCalls    *pSubCalls;

        pSubCalls   = pData->pSubCalls;

        if(uDataReady)											// handle incomeing data from socket
            {
            TcpSubCallRecv(pSubCalls,pData->iSubLevels,uDataReady);
            }

        pSubCalls  += pData->iSubLevels;	                    // goto last subcall
        uDataReady  = pSubCalls->pCountOut(pSubCalls->pHandle); // get bytes to read from buffer
        }

    #endif


return uDataReady;
}

//*****************************************************************************
//*
//*     TcpGetCountSub
//*
//*****************************************************************************/
//  Gets the size of stored data in the receive buffer of the socket. (for subcalls)
//  hHandle : TCP handle from the socket
//  Return the size or 0 if no data is in the buffer, or -1 on an error occurs.
#if SU_HAS_SUBCALLS
static int TcpGetCountSub(TcpHandle hHandle)
{
TcpData        *pData=(TcpData*)hHandle;
unsigned long   uDataReady;


    if(!hHandle)return 0;
    if(ioctlsocket(pData->iSocket,FIONREAD,&uDataReady))        // get bytes to read
        {
        return -1;
        }


return uDataReady;
}
#endif

//*****************************************************************************
//*
//*     TcpIsBrokenEx
//*
//*****************************************************************************/
//  Checks if a tcp connection is brocken or closed and the input buffer.
//  Attention, a closed socket can have data in his queue.
//  hHandle : TCP handle from the socket
//  iMode   : Mode for the handling with the input baffer and the TCP_DATABROKEN flag
//              0 = use the TCP_DATABROKEN flag
//              1 = never check the input buffer
//              2 = check always the input buffer
//  Returns 1 if closed or broken, and 0 if the connection is online.
SUAPI int TcpIsBrokenEx(TcpHandle hHandle,int iMode)
{
TcpData    *pData=(TcpData*)hHandle;
int         iByte;



    if(hHandle==0)return 1;
    if(pData->iSocket==INVALID_SOCKET)return 1;

                                                    // no broken signal if incoming data
    if(iMode==0 && (pData->ucSockFlags&TCP_DATABROKEN))
        {
           iByte = TcpGetCount(hHandle);
        if(iByte>0)return 0;
        }
    else if(iMode==2)                               // always check input puffer
        {
           iByte = TcpGetCount(hHandle);
        if(iByte>0)return 0;
        }


    #if defined(LWIP)
    {
    int     iState;

    iState = lwip_state(pData->iSocket);
                                                    // check the socket state
    if(iState==CLOSED || iState>=FIN_WAIT_1 || iState==SOCKET_ERROR)
        {
        return 1;
        }
    }
    #elif defined(RTTARGET)

    if(!xn_tcp_is_connect(pData->iSocket))          // check the socket state
        {
        return 1;
        }

    #elif defined(SU_WINDOWS)
    {
    WSANETWORKEVENTS    sInfo;

    if(pData->ucConnFlags&FLAG_BROKEN)
        {
        return 1;
        }

    if(WSAEnumNetworkEvents(pData->iSocket,NULL,&sInfo))
        {
        return 1;
        }

    if(sInfo.lNetworkEvents&FD_CLOSE)
        {
        pData->ucConnFlags |= FLAG_BROKEN;
        return 1;
        }
    }
    #elif defined(LINUX)
    {
    struct tcp_info     sInfo;
    socklen_t           iLen;

    iLen = sizeof(sInfo);

    if(getsockopt(pData->iSocket,SOL_TCP,TCP_INFO,(char*)&sInfo,&iLen)==SOCKET_ERROR)
        {
        return 1;
        }

    if(sInfo.tcpi_state==TCP_CLOSE || sInfo.tcpi_state>=TCP_FIN_WAIT1)
        {
        return 1;
        }
    }
    #endif



return 0;
}

//*****************************************************************************
//*
//*     TcpGetLocal
//*
//*****************************************************************************
//  Gets the IP address from the local side of the handle
//  hHandle     : TCP handle of the socket
//  pAddr       : Poiter to the adress memory (0 means not get)
//  pPort       : Pointer fpr the port number (0 means not get)
//  Returns 0 on errors, or 1 if ok
SUAPI int TcpGetLocal(TcpHandle hHandle,IpAddr *pAddr,unsigned short *pPort)
{
TcpData                *pData=(TcpData*)hHandle;
#if SU_USE_IPV6
struct sockaddr_in6_    sAddr;
#else
struct sockaddr_in      sAddr;
#endif
int                     iLen;



    iLen = sizeof(sAddr);

    if(!hHandle || getsockname(pData->iSocket,(struct sockaddr*)&sAddr,(LEN_TYPE)&iLen))
        {
        #if SU_USE_IPV6

        if(pAddr)
            {
            pAddr->aAddr[0]=SU_IPV4_ID;
            pAddr->aAddr[1]=0;
            }

        if(pPort)*pPort=0;

        #else

        if(pAddr)SET_V4_ADDR(pAddr,0);
        if(pPort)*pPort=0;

        #endif

        return 0;
        }



    #if SU_USE_IPV6

    if(sAddr.sin6_family==AF_INET)
        {
        if(pAddr)SET_V4_ADDR(pAddr,((struct sockaddr_in*)&sAddr)->sin_addr.s_addr);
        if(pPort)*pPort = htons   (((struct sockaddr_in*)&sAddr)->sin_port);
        }
    else{
        if(pAddr)memcpy(pAddr->aAddr,sAddr.sin6_addr.s6_addr,16);
        if(pPort)*pPort=htons(sAddr.sin6_port);
        }

    #else

    if(pAddr)SET_V4_ADDR(pAddr,sAddr.sin_addr.s_addr);
    if(pPort)*pPort=htons(sAddr.sin_port);

    #endif



return 1;
}

//*****************************************************************************
//*
//*     IpToString
//*
//*****************************************************************************
//  Converts an ip address to a string
//  pAddr   : ip address (should be nor smaler than 42 bytes)
//  pIp     : pointer to string memory
//  Returns the size of the string.
SUAPI int IpToStringA(IpAddr *pAddr,char *pIp)
{
int     iLen;


    #if SU_USE_IPV6

    if(pAddr->aAddr[0]==SU_IPV4_ID)
        {
        iLen=sprintf(pIp,"%i.%i.%i.%i",pAddr->aByte[4],pAddr->aByte[5],pAddr->aByte[6],pAddr->aByte[7]);
        }
    else if(pAddr->aAddr[2]==SU_IPV4_TAG && !pAddr->aAddr[1] && !pAddr->aAddr[0])
        {
        iLen=sprintf(pIp,"%i.%i.%i.%i",pAddr->aByte[12],pAddr->aByte[13],pAddr->aByte[14],pAddr->aByte[15]);
        }
    else{
        unsigned char  *pData;
        int             iPos;
        int             iMax;
        int             iOff;

        iMax =  0;
        iLen =  0;
        iOff = -1;

        pData = (unsigned char*)(pAddr->aAddr);

        for(iPos=0;iPos<8;iPos++)                   // search for the largest zero area
            {
            if(!pData[iPos*2] && !pData[iPos*2+1])
                {
                iLen++;
                continue;
                }

            if(iLen>iMax)
                {
                iOff=iPos-iLen;
                iMax=iLen;
                }

            iLen=0;
            }

        if(iLen>iMax)
            {
            iOff=iPos-iLen;
            iMax=iLen;
            iLen=0;
            };

        for(iPos=0,iLen=0;iPos<8;iPos++)            // print the values
            {
            if(iPos==iOff)
                {
                if(iPos==0)
                    {
                    pIp[iLen]=':';
                    iLen++;
                    }

                pIp[iLen]=':';iLen++;
                iPos += iMax-1;
                continue;
                }

            if(pData[iPos*2])
                    iLen+=sprintf(pIp+iLen,"%X%02X",pData[iPos*2  ],pData[iPos*2+1]);
            else    iLen+=sprintf(pIp+iLen,"%X"    ,pData[iPos*2+1]);

            if(iPos<7)
                {
                pIp[iLen]=':';
                iLen++;
                }
            }

        pIp[iLen]=0;
        }

    #else

    iLen=sprintf(pIp,"%i.%i.%i.%i",pAddr->aByte[0],pAddr->aByte[1],pAddr->aByte[2],pAddr->aByte[3]);

    #endif


return iLen;
}


//*****************************************************************************
//*
//*     IpToStringW
//*
//*****************************************************************************
//  Converts an ip address to a string
//  pAddr   : ip address (should be nor smaler than 42 bytes)
//  pIp     : pointer to string memory
//  Returns the size of the string.
#if defined(WIN32)
int IpToStringW(IpAddr *pAddr,char *pIp)
{
char    cName[64];
int     iPos;
int     iLen;


    iLen = IpToStringA(pAddr,cName);

    for(iPos=0;iPos<=iLen;iPos++)
        {
        pIp[iPos] = cName[iPos];
        }


return iLen;
}
#endif



//*****************************************************************************
//*
//*     EvnHandleAlloc
//*
//*****************************************************************************/
//  Creates a new event handel
//  Returns the pointer to the handle data
static EvnHandle EvnHandleAlloc()
{
EvnData *pData;


        pData=(EvnData*)MallocHandle(sizeof(EvnData));
    if(!pData)
        {
        return 0;
        }

    pData->hDummy         =  TAG_EVENT;
    pData->iSocket        =  -1;
    pData->ucSignale      =  0;
    pData->ucConnFlags    =  0;


    memset(&pData->sIpAddr,0,sizeof(pData->sIpAddr));



return (EvnHandle)pData;
}


//*****************************************************************************
//*
//*     EvnHandleFree
//*
//*****************************************************************************/
//  Releases a TCP handle
static inline void EvnHandleFree(EvnHandle hHandle)
{

    if(!hHandle)return;

    FreeHandle(hHandle);

}


//*****************************************************************************
//*
//*     EvnOpen
//*
//*****************************************************************************
//  Creates a signal socket for multi wait operations.
//  iSignaled       : is the initial state of the socket
//  Returns the handle or NULL if an error occurs
SUAPI EvnHandle EvnOpen(int iSignaled)
{
EvnHandle       hHandle;
SocketLocal     sLocal;
EvnData        *pData;


    InitSocket();

    if(!SocketAlloc(&sLocal))
        {
        return NULL;
        }


        hHandle = EvnHandleAlloc();
    if(!hHandle)
        {
        SocketClose(&sLocal);
        return 0;
        }

    pData = (EvnData*)hHandle;
    LOCAL_SET(pData,sLocal);

    if(iSignaled)EvnSignal(hHandle);


return hHandle;
}


//*****************************************************************************
//*
//*     EvnClose
//*
//*****************************************************************************
//  Closes an event socket
//  hHandle     : Handle of the TCP socket
//  Returns 0 if an error occurs or 1 if all ok
SUAPI int EvnClose(EvnHandle hHandle)
{
EvnData        *pData=(EvnData*)hHandle;
SocketLocal     sLocal;


    if(!hHandle || !IS_TAG_EVENT(pData->hDummy))
        {
        return 0;
        }

    LOCAL_GET(pData,sLocal);
    SocketClose(&sLocal);
    pData->hDummy = 0;
    EvnHandleFree(hHandle);


return 1;
}


//*****************************************************************************
//*
//*     EvnSignal
//*
//*****************************************************************************
//  Sets a signal socket to the signaled state.
//  hHandle         : is the handle of the socket
SUAPI void EvnSignal(EvnHandle hHandle)
{
EvnData        *pData=(EvnData*)hHandle;


    if(!pData)return;

    pData->ucSignale = 1;
    SocketSignal(PARAM_SIG(pData));

}


//*****************************************************************************
//*
//*     EvnReset
//*
//*****************************************************************************
//  Sets a signal socket to the unsignaled state.
//  hHandle         : is the handle of the socket
SUAPI void EvnReset(EvnHandle hHandle)
{
EvnData        *pData=(EvnData*)hHandle;


    if(!pData)return;

    pData->ucSignale = 0;
    SocketReset(PARAM_SIG(pData));

}

//*****************************************************************************
//*
//*     EvnWait
//*
//*****************************************************************************/
//  Waits until an event object is signaled
//  hHandle     : Is the handle of the event
//                On windows only 64 handles are allowed)
//  iTimeout    : Timeout in millisecounds
//              : Is the offset for the first retuned handle in the array,
//                if more than one is signaled.
//  Returns  -1 on timeout, or the number of the first handle with received data.
SUAPI int EvnWait(EvnHandle hHandle,int iTimeout)
{
EvnData    *pData=(EvnData*)hHandle;



    if(hHandle ==0)return -1;
    if(iTimeout <0)return  0;

    if(pData->ucSignale)return 1;


//******************** Windows ************************************************


    #ifdef SU_WINDOWS

	int    iRet;
	                                                        // wait for the next event
    iRet = WSAWaitForMultipleEvents(1,&pData->hEvent,FALSE,iTimeout,FALSE);

    if(iRet!=WSA_WAIT_EVENT_0)return 0;


//******************** select *************************************************


    #else

    struct timeval  sTimeVal;
    fd_set          sFdRead;

    sTimeVal.tv_sec  =  iTimeout/1000;
    sTimeVal.tv_usec = (iTimeout%1000)*1000;

    FD_ZERO(&sFdRead);
    FD_SET(pData->iSocket,&sFdRead);


    if(select(1,&sFdRead,0,0,&sTimeVal)<=0)             // wait until data received
        {
        return 0;
        }

    #endif



return 1;
}


//*****************************************************************************
//*
//*     EvnWaitMulti
//*
//*****************************************************************************/
//  Waits until the next data from a socket array was received.
//  The array could contain TCP,UDP and event sockets.
//  pHandle     : Pointer to the handle array of the TCP sockets
//                On windows only 64 handles are allowed)
//  iCount      : Is the count of handles in the array (0..n)
//                If this value is zero the function sleeps iTimeout ms.
//  iTimeout    : Timeout in millisecounds
//              : Is the offset for the first retuned handle in the array,
//                if more than one is signaled.
//  iOffset     : Ist the first offset in the array, for the begin
//                of seaching signaled sockets.
//  Returns  -1 on timeout, or the number of the first handle with received data.
SUAPI int EvnWaitMulti(void *pHandle,int iCount,int iTimeout,int iOffset)
{
#if SU_USE_IPV6
EvnData        *aServers[16];
int             aSockIds[16];
int             iAdd=0;
int             iOff;
#endif
EvnData       **pData=(EvnData**)pHandle;
int             iDone;
int             iPos;



    if(pHandle==0)return -1;
    if(iTimeout<0 || iCount<0)
        {
        return -1;
        }

    if(iCount==0)
        {
        SysSleep(iTimeout);
        return -1;
        }

    #if SU_HAS_SUBCALLS

    TcpData    *pPointer;
    unsigned    uEndTime;
    unsigned    uIndex;

    uIndex = iOffset;

    for(iPos=0;iPos<iCount;iPos++,uIndex++)             // search for signaled handles
        {
        if(uIndex>=(unsigned)iCount)uIndex=0;

        pPointer = (TcpData*)pData[uIndex];

        if(IS_TAG_TCP(pPointer->hServer))
            {
            iDone = TcpGetCount((TcpHandle)pPointer);
            }
        else if(IS_TAG_SERVER(pPointer->hServer))
            {
            iDone = TcpWaitServer((TcpHandle)pPointer,0);
            }
        else if(IS_TAG_UDP(pPointer->hServer))
            {
            iDone = UdpGetCount((UdpHandle)pPointer);
            }
        else if(IS_TAG_EVENT(pPointer->hServer))
            {
            iDone = EvnWait((EvnHandle)pPointer,0);
            }
        else{
            iDone = 0;
            }

        if(iDone>0)return uIndex;
        }

    if(iTimeout==0)return -1;

    uEndTime = SysTickCount()+iTimeout;

Restart:

    #endif


//******************** Windows ************************************************

    #ifdef SU_WINDOWS

    int                 aSocket[WSA_MAXIMUM_WAIT_EVENTS];
    DWORD               aSockId[WSA_MAXIMUM_WAIT_EVENTS];
    WSAEVENT            aEvents[WSA_MAXIMUM_WAIT_EVENTS];
    WSANETWORKEVENTS    sInfo;
    unsigned long       uByte;
    unsigned            uStop;
    EvnData            *pEntry;
    void               *hDummy;
    int                 iLast;
    int                 iIdx;
    int                 iAll;

    if(iCount>WSA_MAXIMUM_WAIT_EVENTS)
        {
        iCount = WSA_MAXIMUM_WAIT_EVENTS;
        }

    if(iOffset<0 || iOffset>=iCount)
        {
        iOffset = 0;
        }

    iLast = iOffset;
    iDone = -1;
    iAll  =  0;

    for(iPos=0;iPos<iCount;iPos++)                      // set all sockets
        {
        if(!pData[iPos])
            {
            iLast--;                                    // invalid handle data
            continue;
            }

        hDummy        = pData[iPos]->hDummy;
        aEvents[iAll] = pData[iPos]->hEvent;
        aSockId[iAll] = iPos;

        if(IS_TAG_EVENT(hDummy))                        // is it an event socket
            {
            aSocket[iAll] = -1;
            iAll++;
            }
        else{
            aSocket[iAll] = pData[iPos]->iSocket;

            if(IS_TAG_SERVER(hDummy))                   // is it a sever socked
                {
                iAll++;
                }
            else{
                if(pData[iPos]->ucConnFlags&FLAG_BROKEN)// is the connection broken
                    {
                    if((iDone<iOffset && iPos>=iOffset) || iDone<0)iDone=iPos;
                    WSASetEvent(aEvents[iAll]);
                    iTimeout = 0;
                    iAll++;
                    }                                   // already data received
                else if(!ioctlsocket(aSocket[iPos],FIONREAD,&uByte) && uByte)
                    {
                    if((iDone<iOffset && iPos>=iOffset) || iDone<0)iDone=iPos;
                    iTimeout = 0;
                    iLast--;
                    }
                else{
                    iAll++;
                    }
                }
            }

        #if SU_USE_IPV6
                                                        // is it a multiple server port
        if(!IS_TAG_MULTI(hDummy) || iAdd>=16)
            {
            continue;
            }

        aServers[iAdd] = (EvnData*)hDummy;
        aSockIds[iAdd] = iPos;
        iAdd++;

        #endif
        }

    #if SU_USE_IPV6

    if(iAll+iAdd>WSA_MAXIMUM_WAIT_EVENTS)               // check count off signals
        {
        iAdd = WSA_MAXIMUM_WAIT_EVENTS-iAll;
        }

    for(iOff=0;iOff<iAdd;iOff++)                        // add multiple server sockets
        {
        aEvents[iAll] = aServers[iOff]->hEvent;
        aSocket[iAll] = aServers[iOff]->iSocket;
        aSockId[iAll] = aSockIds[iOff];
        iAll++;
        }

    #endif

    uStop = GetTickCount()+iTimeout;

Retry:

                                                        // wait for the next event
    iPos = WSAWaitForMultipleEvents(iAll,aEvents,FALSE,iTimeout,FALSE);

                                                        // is an event signaled
    if(iPos<(int)WSA_WAIT_EVENT_0+iAll && iPos>=(int)WSA_WAIT_EVENT_0)
        {
        iPos -= WSA_WAIT_EVENT_0;

        if(iLast>iPos)                                  // is an event in the upper part signaled
            {
            iIdx = WSAWaitForMultipleEvents(iAll-iLast,aEvents+iLast,FALSE,0,FALSE);

            if(iIdx<(int)WSA_WAIT_EVENT_0+iAll && iIdx>=(int)WSA_WAIT_EVENT_0)
                {
                iIdx -= WSA_WAIT_EVENT_0;
                iPos  = iLast+iIdx;
                }
            }

        iIdx = aSockId[iPos];                           // real index in request array

        #if SU_USE_IPV6

        iAll -= iAdd;

        #endif

                                                        // reset the signaled events
        if(aSocket[iPos]>=0 && !WSAEnumNetworkEvents(aSocket[iPos],NULL,&sInfo))
            {
            #if SU_USE_IPV6

            if(iPos<iAll)
                    pEntry = pData   [iIdx     ];
            else    pEntry = aServers[iPos-iAll];

            #else

            pEntry = pData[iIdx];

            #endif

            pEntry->ucConnFlags |= (sInfo.lNetworkEvents&~FLAG_DOWN);


            if(sInfo.lNetworkEvents&(FD_ACCEPT|FD_WRITE|FD_READ) && !(pEntry->ucConnFlags&FD_CLOSE))
                {
                WSAResetEvent(pEntry->hEvent);
														// check are there bytes after a receive event 
				if(sInfo.lNetworkEvents==FD_READ && IS_TAG_CLIENT(pEntry->hDummy))
				if(!ioctlsocket(aSocket[iPos],FIONREAD,&uByte) && uByte)
					{
					pEntry->ucConnFlags &= ~FD_READ;
					}
                }
                                                        // if no read signal try it again
            if(iDone<0 && !(pEntry->ucConnFlags&(FD_ACCEPT|FD_CLOSE|FD_READ)))
                {
                   iTimeout = uStop-GetTickCount();
                if(iTimeout>0)
                    {
					WSAResetEvent(aEvents[iPos]);

                    #if SU_USE_IPV6
                    iAll += iAdd;
                    #endif
                    goto Retry;
                    }
                }
            }

        #if SU_USE_IPV6

        if(iPos>=iAll)									// is the slot from a multiple server port
            {
            iIdx = aSockIds[iPos-iAll];
            }

        #endif

        if(iDone>=0 && iDone<iIdx && iDone>=iOffset)	// was a lower event happened
            {
            goto Exit;
            }

        iDone = iIdx;
        goto Exit;
        }

    if(iDone>=0)                                        // was a lower event happened
        {
        goto Exit;
        }


//******************** select *************************************************


    #else

    struct timeval  sTimeVal;
    fd_set          sFdRead;
    int             iMax=1;

    sTimeVal.tv_sec  =  iTimeout/1000;
    sTimeVal.tv_usec = (iTimeout%1000)*1000;

    FD_ZERO(&sFdRead);


    for(iPos=0;iPos<iCount;iPos++)                      // set all sockets
        {
        if(!pData[iPos])
            {
            continue;
            }

        if(iMax<=(int)pData[iPos]->iSocket)
            {
            iMax = pData[iPos]->iSocket+1;
            }

        FD_SET(pData[iPos]->iSocket,&sFdRead);

        #if SU_USE_IPV6
                                                        // is it a multiple server port
        if(!IS_TAG_MULTI(pData[iPos]->hDummy) || iAdd>=16)
            {
            continue;
            }

        aServers[iAdd] = (EvnData*)pData[iPos]->hDummy;
        aSockIds[iAdd] = iPos;
        iAdd++;

        #endif
        }


    #if SU_USE_IPV6

    for(iOff=0;iOff<iAdd;iOff++)                        // add dual server sockets
        {
        if(iMax<=(int)aServers[iOff]->iSocket)
            {
            iMax = aServers[iOff]->iSocket+1;
            }

        FD_SET(aServers[iOff]->iSocket,&sFdRead);

        iCount++;
        }

    #endif


    if(select(iMax,&sFdRead,0,0,&sTimeVal)<=0)          // wait until data received
        {
        return -1;
        }

    if(iOffset<0)iOffset=0;


    for(iPos=0;iPos<iCount;iPos++,iOffset++)            // find signaled socket
        {
        if(iOffset>=iCount)iOffset=0;

        if(FD_ISSET(pData[iOffset]->iSocket,&sFdRead))
            {
            #if SU_USE_IPV6

            iCount -= iAdd;

            if(iOffset>=iCount)                         // is the slot from a multiple server port
                {
                iOffset = aSockIds[iOffset-iCount];
                }

            #endif

            iDone = iOffset;
            goto Exit;
            }
        }

    #endif

    return -1;

Exit:


    #if SU_HAS_SUBCALLS

    pPointer = (TcpData*)pData[iDone];
                                                        // data receivet on subcall stream
    if(IS_TAG_TCP(pPointer->hServer) && pPointer->pSubCalls && !TcpGetCount((TcpHandle)pPointer))
        {
           iTimeout = uEndTime-SysTickCount();
        if(iTimeout<0)iTimeout=0;

        goto Restart;
        }

    #endif


return iDone;
}


//*****************************************************************************
//*
//*     TcpSubCallRecv
//*
//*****************************************************************************
//  Handles a read access over all subcalls
//  pSubCalls   : pointer to the subcall array
//	iLevels		: is the count of subcalls in the chain
//  iBytes      : count of byte do read from the base socket
//                -1 means get the byte count from tcp level
//  Returns -1 in an error occurs or 0 if all ok
#if SU_HAS_SUBCALLS
static int TcpSubCallRecv(TcpSubCalls *pSubCalls,int iLevels,int iBytes)
{
unsigned char   aBuffer[SUB_BUFFER];
TcpSubCalls    *pCall;
TcpSubCalls    *pNext;
int             iSize;
int             iFlag;
int             iAdd;
int             iPos;



    if(iBytes==-1)                                      // get byte count from tcp level
        {
           iBytes = pSubCalls->pCountOut(pSubCalls->pHandle);
        if(iBytes<0)iBytes=0;
        }


    for(iFlag=iBytes;iFlag;iBytes-=iAdd)
        {
        iAdd  = (iBytes<sizeof(aBuffer))? iBytes:sizeof(aBuffer);
        pCall = pSubCalls;
        iSize = iAdd;
        iFlag = iAdd;


        for(iPos=0;iPos<iLevels;iPos++,pCall=pNext)		// receive data throw all levels (up)
            {
            if(iPos)
                {
                iSize = pCall->pCountOut(pCall->pHandle);
                }

            pNext = pCall+1;

            if(iSize)
                {
                if(iSize<0)goto Error;
                if(iSize>sizeof(aBuffer))iSize=sizeof(aBuffer);
                   iSize = pCall->pOutRd(pCall->pHandle,aBuffer,iSize);
                if(iSize<0)goto Error;
                   iSize = pNext->pInpWr(pNext->pHandle,aBuffer,iSize);
                if(iSize<0)goto Error;

                iFlag = 1;
                }
            }


        for(iPos--;iPos>=0;iPos--,pNext=pCall)          // send data throw all levels (down)
            {
            pCall = pNext-1;

               iSize = pNext->pCountInp(pNext->pHandle);
            if(iSize)
                {
                if(iSize<0)goto Error;
                if(iSize>sizeof(aBuffer))iSize=sizeof(aBuffer);
                   iSize = pNext->pInpRd(pNext->pHandle,aBuffer,iSize);
                if(iSize<0)goto Error;
                   iSize = pCall->pOutWr(pCall->pHandle,aBuffer,iSize);
                if(iSize<0)goto Error;

                iFlag = 1;
                }
            }
        }


    return 0;

Error:

    return -1;
}
#endif

//*****************************************************************************
//*
//*     TcpSubCallSend
//*
//*****************************************************************************
//  Handles a write access over all subcalls
//  pSubCalls   : pointer to the subcall array
//	iLevels		: is the count of subcalls in the chain
//  iBytes      : count of byte do read from the base socket
//                -1 means check chain
//  pData       : pointer to the data for write
//  Returns -1 in an error occurs or the transmitted bytes.
#if SU_HAS_SUBCALLS
static int TcpSubCallSend(TcpSubCalls *pSubCalls,int iLevels,int iBytes,const char *pData)
{
unsigned char   aBuffer[SUB_BUFFER];
TcpSubCalls    *pCall;
TcpSubCalls    *pNext;
int             iSize;
int             iFlag;
int             iRet;
int             iPos;



    iRet = 0;


    for(;;)
        {
        pNext = pSubCalls+iLevels;
        iSize = iBytes;
		iFlag = iBytes;


        if(iBytes>0)                                    // write input bytes to stream
            {
               iSize = pNext->pOutWr(pNext->pHandle,pData,iBytes);
            if(iSize<0)goto Error;

            iRet   += iSize;
            pData  += iSize;
            iBytes -= iSize;
            }


        for(iPos=iLevels-1;iPos>=0;iPos--,pNext=pCall)  // send data throw all levels (down)
            {
            pCall = pNext-1;

               iSize = pNext->pCountInp(pNext->pHandle);
            if(iSize)
                {
                if(iSize<0)goto Error;
                if(iSize>sizeof(aBuffer))iSize=sizeof(aBuffer);
                   iSize = pNext->pInpRd(pNext->pHandle,aBuffer,iSize);
                if(iSize<0)goto Error;
                   iSize = pCall->pOutWr(pCall->pHandle,aBuffer,iSize);
                if(iSize<0)goto Error;

                iFlag = 1;
                }
            }


        for(iPos=0;iPos<iLevels;iPos++,pCall=pNext)     // receive data throw all levels (up)
            {
            pNext = pCall+1;

               iSize = pCall->pCountOut(pCall->pHandle);
            if(iSize)
                {
                if(iSize<0)goto Error;
                if(iSize>sizeof(aBuffer))iSize=sizeof(aBuffer);
                   iSize = pCall->pOutRd(pCall->pHandle,aBuffer,iSize);
                if(iSize<0)goto Error;
                   iSize = pNext->pInpWr(pNext->pHandle,aBuffer,iSize);
                if(iSize<0)goto Error;

                iFlag = 1;
                }
            }

		if(iFlag<=0)									// end of transmission
			{
			break;
			}
		}


    return iRet;

Error:

    return -1;
}
#endif

//*****************************************************************************
//*
//*		TcpSubFlash
//*
//*****************************************************************************
//  Flushes the output data on a socket with sub calls.
//  hHandle     : TCP handle of the socket.
//	returns 1 if the socke has sub calls, and 0 if not, or -1 if an error occurs.
int TcpSubFlash(TcpHandle hHandle)
{
TcpData	   *pData=(TcpData*)hHandle;



	if(!pData)return -1;

	#if SU_HAS_SUBCALLS

	int			iRet;

	if(pData->pSubCalls)
		{
		   iRet = TcpSubCallSend(pData->pSubCalls,pData->iSubLevels);
		if(iRet<0)return -1;

		return 1;
		}

	#endif


return 0;
}

//*****************************************************************************
//*
//*     TcpAddSubCall
//*
//*****************************************************************************
//  Inserts a subcall entry in a tcp handle.
//  hHandle     : TCP handle of the socket.
//  pNewCall    : contains the subcall functions.
//  pHandle     : is the handle for the input side of this subcall.
//                if this is zero use the handle from pNewCall->pHandle.
//  Returns the count of all subcalls or 0 if an erro occurs
SUAPI int TcpAddSubCall(TcpHandle hHandle,const TcpSubCalls *pNewCall,void *pHandle)
{
#if SU_HAS_SUBCALLS == 0
return 0;
#else
TcpSubCalls    *pSubCalls;
TcpData        *pData;
unsigned        uLevel;




        pData = (TcpData*)hHandle;
    if(!pData || !IS_TAG_TCP(pData->hServer))           // check tcp handle
        {
        return 0;
        }


        pSubCalls = pData->pSubCalls;
    if(!pSubCalls)                                      // initalize subcall array
        {
            pSubCalls = (TcpSubCalls*)MallocHandle(sizeof(TcpSubCalls)*(SU_HAS_SUBCALLS+1));
        if(!pSubCalls)return 0;

		pData->iSubLevels		= 0;
        pData->pSubCalls        = pSubCalls;
        pSubCalls->pHandle      = pData;
        pSubCalls->pOutRd       = (SubRecv)TcpGetSub;
        pSubCalls->pOutWr       = (SubSend)TcpPutSub;
        pSubCalls->pCountOut    = (SubCall)TcpGetCountSub;
        pSubCalls->pInpRd       = 0;
        pSubCalls->pInpWr       = 0;
        pSubCalls->pCountInp    = 0;
        pSubCalls->pClose       = 0;
        uLevel                  = 1;
        }
    else{
           uLevel = pData->iSubLevels+1;
        if(uLevel>=SU_HAS_SUBCALLS)                     // too much subcalls attached
            {
            return 0;
            }
        }


	pSubCalls += uLevel; 								// insert new sub call
    memcpy(pSubCalls,pNewCall,sizeof(TcpSubCalls));
    if(pHandle)pSubCalls->pHandle = pHandle;

	pData->iSubLevels = uLevel;


return uLevel;
#endif
}

